"""Data Modules."""

from .generic_classification_module import (  # noqa: F401
    OlindaCDataModule,
    OlindaCDataset,
)
from .generic_regression_module import OlindaRDataModule, OlindaRDataset  # noqa: F401



"""Reusable BLoCs."""

from http import HTTPStatus
from typing import Any, Union

from flask import Blueprint, make_response, request, Response
from pydantic import parse_raw_as, ValidationError
from pytorch_lightning import LightningModule
import tenseal as ts
from torch import nn

from chemxor.schema.fhe_model import (
    ModelContextParams,
    ModelInfo,
    PartFHEModelQueryGetResponse,
    PartFHEModelQueryPostRequest,
    PartFHEModelQueryPostResponse,
)


def process_fhe_model_query(
    query: PartFHEModelQueryPostRequest,
    model: Union[nn.Module, LightningModule],
) -> Any:
    """Evaluate fhe model."""
    # Parse context from request
    context = ts.context_from(bytes.fromhex(query.ts_context))

    # convert enc input to enc vector
    enc_input = ts.ckks_vector_from(context, bytes.fromhex(query.model_input))
    return model(enc_input, query.model_step)


def generate_blueprint(
    part_net: Union[nn.Module, LightningModule],
    name: str = "partnet",
    url_prefix: str = "/v1/fhe",
) -> Blueprint:
    """Generate flask blueprint for a specific model.

    Args:
        part_net (Union[nn.Module, LightningModule]): FHE model
        name (str): Name of blueprint
        url_prefix (str): Blueprint url_prefix

    Returns:
        Blueprint: Flask blueprint
    """
    part_net_blueprint = Blueprint(name, name, url_prefix=url_prefix)

    @part_net_blueprint.post("/")
    def part_net_model_post() -> Response:
        """Partnet POST request handler.

        Returns:
            Response: Flask response.
        """
        # Parse json from requests
        try:
            model_query_request = parse_raw_as(
                PartFHEModelQueryPostRequest, request.json
            )
        except ValidationError:
            response_body = PartFHEModelQueryPostResponse()
            return make_response(response_body.json(), HTTPStatus.BAD_REQUEST)
        except Exception:
            response_body = PartFHEModelQueryPostResponse()
            return make_response(response_body.json(), HTTPStatus.INTERNAL_SERVER_ERROR)

        enc_output_tensor = process_fhe_model_query(model_query_request, part_net)

        # determine next step
        next_step = (
            (model_query_request.model_step + 1)
            if (model_query_request.model_step + 1) <= part_net.steps
            else None
        )
        response_body = PartFHEModelQueryPostResponse(
            model_output=enc_output_tensor.serialize().hex(),
            next_step=next_step,
            preprocess_next_args=part_net.pre_process[model_query_request.model_step],
        )
        return make_response(response_body.json(), HTTPStatus.OK)

    @part_net_blueprint.get("/")
    def part_net_model_get() -> Response:
        """Partnet GET request handler.

        Returns:
            Response: Flask response.
        """
        response_body = PartFHEModelQueryGetResponse(
            model_info=ModelInfo(
                model_name=str(part_net),
                model_steps=part_net.steps,
                context_params=ModelContextParams(
                    bit_scale=part_net.bits_scale,
                    poly_modulus_degree=part_net.poly_modulus_degree,
                    coeff_mod_bit_sizes=part_net.coeff_mod_bit_sizes,
                ),
            )
        )
        return make_response(response_body.json(), HTTPStatus.OK)

    return part_net_blueprint



"""Encrypted Convolution Dataset wrapper."""

from typing import Any

import tenseal as ts
from torch.utils.data import Dataset


class EncConvDataset(Dataset):
    """Encrypted Convolution Dataset."""

    def __init__(
        self: "EncConvDataset",
        context: ts.Context,
        dataset: Dataset,
        kernel_shape: tuple,
        stride: int,
        input_shape: tuple = (28, 28),
    ) -> None:
        """Encrypted dataset.

        Args:
            context (ts.Context): Tenseal encryption context.
            dataset (Dataset): Initialized dataset class to wrap.
            kernel_shape (tuple): Shape of the convolution kernel.
            stride (int): Stride length.
            input_shape (tuple): Shape of the input image. Defaults to (28, 28).

        """
        self.context = context
        self.dataset = dataset
        self.kernel_shape = kernel_shape
        self.stride = stride
        self.input_shape = input_shape

    def __len__(self: "EncConvDataset") -> int:
        """Length of the dataset.

        Returns:
            int: Length of the dataset.
        """
        return self.dataset.__len__()

    def __getitem__(self: "EncConvDataset", index: int) -> Any:
        """Get item from the dataset.

        Args:
            index (int): index of the item.

        Returns:
            Any: Item
        """
        items = self.dataset.__getitem__(index)
        # Encrypt items
        enc_x, windows_nb = ts.im2col_encoding(
            self.context,
            items[0].view(self.input_shape[0], self.input_shape[1]).tolist(),
            self.kernel_shape[0],
            self.kernel_shape[1],
            self.stride,
        )
        enc_y = ts.ckks_tensor(self.context, [items[1]])
        return tuple([enc_x, enc_y, windows_nb])



"""OSM data module."""

from copy import deepcopy
from pathlib import Path
from typing import Any, Optional

import pandas as pd
import pytorch_lightning as pl
import tenseal as ts
import torch as t
from torch.utils.data import DataLoader, Dataset, random_split

from chemxor.data.enc_dataset import EncDataset


class OSMDataset(Dataset):
    """OSM Dataset."""

    def __init__(
        self: "OSMDataset",
        train_set_dir_path: Path = (
            Path(__file__)  # noqa: B008
            .parents[3]
            .joinpath("data/01_raw/train_set.csv")
        ),
        train_res_dir_path: Path = (
            Path(__file__)  # noqa: B008
            .parents[3]
            .joinpath("data/01_raw/train_res.csv")
        ),
        transform: Optional[Any] = None,
        target_transform: Optional[Any] = None,
    ) -> None:
        """OSM dataset.

        Args:
            train_set_dir_path (Path): train set path.
                Defaults to Path(__file__).parents[1].joinpath("/data/01_raw/train_set.csv").
            train_res_dir_path (Path): train res path.
                Defaults to Path(__file__).parents[1].joinpath("/data/01_raw/train_res.csv").
            transform (Optional[Any]): Tranforms for the inputs. Defaults to None.
            target_transform (Optional[Any]): Transforms for the target. Defaults to None.

        """
        self.train_set_df = pd.read_csv(train_set_dir_path.absolute())
        self.train_res_df = pd.read_csv(train_res_dir_path.absolute())
        self.transform = transform
        self.target_transform = target_transform

    def __len__(self: "OSMDataset") -> int:
        """Length of the dataset.

        Returns:
            int: Length of the dataset.
        """
        return len(self.train_set_df)

    def __getitem__(self: "OSMDataset", index: int) -> Any:
        """Get item from the dataset.

        Args:
            index (int): index of the item.

        Returns:
            Any: Item
        """
        molecule = t.tensor(self.train_set_df.iloc[index][2:], dtype=t.float32)
        label = self.train_res_df.iloc[index][2]
        if label < 1:
            label = t.tensor([1, 0], dtype=t.float32)
        else:
            label = t.tensor([0, 1], dtype=t.float32)
        if self.transform:
            molecule = self.transform(molecule)
        if self.target_transform:
            label = self.target_transform(label)
        return molecule, label


class OSMDataModule(pl.LightningDataModule):
    """OSM data module."""

    def __init__(
        self: "OSMDataModule",
        train_set_dir_path: Path = (
            Path(__file__)  # noqa: B008
            .parents[3]
            .joinpath("data/01_raw/train_set.csv")
        ),
        train_res_dir_path: Path = (
            Path(__file__)  # noqa: B008
            .parents[3]
            .joinpath("data/01_raw/train_res.csv")
        ),
        batch_size: int = 10,
        transform: Optional[Any] = None,
        target_transform: Optional[Any] = None,
    ) -> None:
        """OSM data module init.

        Args:
            train_set_dir_path (Path): train set path.
                Defaults to Path(__file__).parents[1].joinpath("/data/01_raw/train_set.csv").
            train_res_dir_path (Path): train res path.
                Defaults to Path(__file__).parents[1].joinpath("/data/01_raw/train_res.csv").
            batch_size (int): batch size. Defaults to 32.
            transform (Optional[Any]): Tranforms for the inputs. Defaults to None.
            target_transform (Optional[Any]): Transforms for the target. Defaults to None.
        """
        super().__init__()
        self.train_set_dir_path = train_set_dir_path
        self.train_res_dir_path = train_res_dir_path
        self.batch_size = batch_size
        self.transform = transform
        self.target_transform = target_transform

    def prepare_data(self: "OSMDataModule") -> None:
        """Prepare data."""
        # download, split, etc...
        # only called on 1 GPU/TPU in distributed
        pass

    def setup(self: "OSMDataModule", stage: Optional[str]) -> None:
        """Setup dataloaders.

        Args:
            stage (Optional[str]): Optional pipeline state
        """
        osm_full = OSMDataset(
            self.train_set_dir_path,
            self.train_res_dir_path,
            self.transform,
            self.target_transform,
        )
        self.osm_train, self.osm_val = random_split(
            osm_full, [270, 117], t.Generator().manual_seed(7777)
        )

        # hack for now, please fixme later
        self.enc_osm_train = deepcopy(self.osm_train)
        self.osm_test = deepcopy(self.osm_val)
        self.enc_osm_test = deepcopy(self.osm_val)
        self.osm_predict = deepcopy(self.osm_val)
        self.enc_osm_predict = deepcopy(self.osm_val)

        # # split train into train and validation
        # self.osm_train, self.osm_val = random_split(
        #     self.osm_train,
        #     [
        #         int(len(self.osm_train) * 80),
        #         len(self.osm_train) - int(len(self.osm_train) * 80),
        #     ],
        # )
        # # split test into test and predict
        # self.osm_test, self.osm_predict = random_split(
        #     self.osm_test,
        #     [
        #         int(len(self.osm_test) * 80),
        #         len(self.osm_test) - int(len(self.osm_test) * 80),
        #     ],
        # )

    def train_dataloader(self: "OSMDataModule") -> DataLoader:
        """Train dataloader.

        Returns:
            DataLoader: train dataloader
        """
        return DataLoader(self.osm_train, batch_size=self.batch_size)

    def enc_train_dataloader(self: "OSMDataModule", context: ts.Context) -> DataLoader:
        """Encrypted train dataloader.

        Args:
            context (ts.Context): Tenseal encryption context.

        Returns:
            DataLoader: train dataloader
        """
        enc_osm_train = EncDataset(context, self.enc_osm_train)
        return DataLoader(enc_osm_train, batch_size=None)

    def val_dataloader(self: "OSMDataModule") -> DataLoader:
        """Val dataloader.

        Returns:
            DataLoader: val dataloader
        """
        return DataLoader(self.osm_val, batch_size=self.batch_size)

    def enc_val_dataloader(self: "OSMDataModule", context: ts.Context) -> DataLoader:
        """Encrypted val dataloader.

        Args:
            context (ts.Context): Tenseal encryption context.

        Returns:
            DataLoader: val dataloader
        """
        enc_osm_val = EncDataset(context, self.enc_osm_val)
        return DataLoader(enc_osm_val, batch_size=None)

    def test_dataloader(self: "OSMDataModule") -> DataLoader:
        """Test data loader.

        Returns:
            DataLoader : test dataloader
        """
        return DataLoader(self.osm_test, batch_size=self.batch_size)

    def enc_test_dataloader(self: "OSMDataModule", context: ts.Context) -> DataLoader:
        """Encrypted test dataloader.

        Args:
            context (ts.Context): Tenseal encryption context.

        Returns:
            DataLoader: test dataloader
        """
        enc_osm_test = EncDataset(context, self.enc_osm_test)
        return DataLoader(enc_osm_test, batch_size=None)

    def predict_dataloader(self: "OSMDataModule") -> DataLoader:
        """Predict data loader.

        Returns:
            DataLoader : predict dataloader
        """
        return DataLoader(self.osm_predict, batch_size=self.batch_size)

    def enc_predict_dataloader(
        self: "OSMDataModule", context: ts.Context
    ) -> DataLoader:
        """Encrypted predict dataloader.

        Args:
            context (ts.Context): Tenseal encryption context.

        Returns:
            DataLoader: predict dataloader
        """
        enc_osm_predict = EncDataset(context, self.enc_osm_predict)
        return DataLoader(enc_osm_predict, batch_size=None)

    def teardown(self: "OSMDataModule", stage: Optional[str] = None) -> None:
        """Teardown of data module."""
        pass



"""Partitioned network Service."""

from typing import Optional, Union

from flask import Blueprint, Flask
from pytorch_lightning import LightningModule
from torch import nn

from chemxor.service.blocs import generate_blueprint


class PartitionNetServer:
    """Partitioned network service."""

    def __init__(
        self: "PartitionNetServer",
        part_net: Optional[Union[nn.Module, LightningModule]] = None,
        flask_app: Optional[Flask] = None,
    ) -> None:
        """Init."""
        self.part_net = part_net
        self.flask_app = flask_app or Flask(str(part_net))
        self.flask_bp = generate_blueprint(self.part_net)
        self.flask_app.register_blueprint(self.flask_bp)

    def get_flask_app(self: "PartitionNetServer") -> Optional[Flask]:
        """Return Flask app if present."""
        return self.flask_app

    def get_flask_blueprint(self: "PartitionNetServer") -> Optional[Blueprint]:
        """Return Flask blueprint if present."""
        return self.flask_bp

    def run(self: "PartitionNetServer") -> None:
        """Start flask server."""
        self.flask_app.run()



"""Cryptic Sage route."""

from flask import Blueprint
from flask import Response

cryptic_sage_bp = Blueprint("cryptic_sage", __name__, url_prefix="/v1/crypticsage")


@cryptic_sage_bp.post("/")
def cryptic_sage_post() -> Response:
    """Test encryption.

    Returns:
        Response: Flask response
    """
    return "OK"



"""Project pipelines."""
from typing import Dict

from kedro.pipeline import Pipeline

from chemxor.pipelines.olinda_train_classification import olinda_one_cls_pipeline
from chemxor.pipelines.olinda_train_regression import olinda_one_reg_pipeline
from chemxor.pipelines.smiles_preprocess import convert_smiles_to_imgs_pipeline


def register_pipelines() -> Dict[str, Pipeline]:
    """Register the project's pipelines.

    Returns:
        A mapping from a pipeline name to a ``Pipeline`` object.
    """
    return {
        "olinda_one_reg": olinda_one_reg_pipeline,
        "olinda_one_cls": olinda_one_cls_pipeline,
        "preprocess_smiles": convert_smiles_to_imgs_pipeline,
        "__default__": olinda_one_reg_pipeline,
    }



"""SmilesImg data module."""

from glob import glob
import math
from pathlib import Path
import re
from typing import Any, Optional, Union

import pandas as pd
from pyarrow import feather
import pytorch_lightning as pl
import tenseal as ts
import torch as t
from torch import nn
from torch.utils.data import DataLoader, Dataset, random_split

from chemxor.data_modules.enc_conv_dataset import EncConvDataset
from chemxor.utils import get_project_root_path

project_root_path = get_project_root_path()
default_path = project_root_path.joinpath(
    "data/01_raw/ModelPreds/eos2r5a/ersilia_output.csv"
)


class SmilesImgDataset(Dataset):
    """SmilesImg Dataset."""

    def __init__(
        self: "SmilesImgDataset",
        preds_csv_path: Path = default_path,
        imgs_df_dir: Path = default_path.parents[0],
        transform: Optional[Any] = None,
        target_transform: Optional[Any] = None,
    ) -> None:
        """Smiles as image dataset.

        Args:
            preds_csv_path (Path): preds csv path.
                Defaults to default_path.
            imgs_df_dir (Path): images csv path.
                Defaults to default_path.parents[0].
            transform (Optional[Any]): Tranforms for the inputs. Defaults to None.
            target_transform (Optional[Any]): Transforms for the target. Defaults to None.

        """
        self.preds_df_full = pd.read_csv(preds_csv_path.absolute())
        self.imgs_dfs_list = [
            path for path in glob(f"{imgs_df_dir.absolute()}/sm_to_imgs_*")
        ]
        self.imgs_dfs_list = sorted(self.imgs_dfs_list, key=self._get_order)
        self.imgs_dfs_len = len(self.imgs_dfs_list)
        self.df_len = len(self.preds_df_full)
        self.splits = [
            x for x in range(0, self.df_len + 1, (self.df_len + 1) // self.imgs_dfs_len)
        ]
        self.transform = transform
        self.target_transform = target_transform

    def _get_order(self: "SmilesImgDataset", file: str) -> int:
        file_pattern = re.compile(r".*?(\d+).*?")
        match = file_pattern.match(Path(file).name)
        if not match:
            return math.inf
        return int(match.groups()[0])

    def __len__(self: "SmilesImgDataset") -> int:
        """Length of the dataset.

        Returns:
            int: Length of the dataset.
        """
        return len(self.preds_df_full) - 30

    def __getitem__(self: "SmilesImgDataset", index: int) -> Any:
        """Get item from the dataset.

        Args:
            index (int): index of the item.

        Returns:
            Any: Item
        """
        # determine partition for index
        partition = 0
        for i, split in enumerate(self.splits):
            if index < split:
                partition = i - 1
                break
        # determine chunk of the partition
        df = feather.read_feather(self.imgs_dfs_list[partition]).set_index(
            "global_index"
        )
        # Extract smile string
        molecule_img = t.tensor(
            df.loc[index, [str(x) for x in range(0, 1024)]],
            dtype=t.float,
        ).reshape(1, 32, 32)

        # Extract target
        label = t.tensor(df.loc[index, ["target"]], dtype=t.float)
        # Apply input transforms
        if self.transform:
            molecule_img = self.transform(molecule_img)

        # Apply target transforms
        if self.target_transform:
            label = self.target_transform(label)

        # Return input and target
        return molecule_img, label


class SmilesImgDataModule(pl.LightningDataModule):
    """SmilesImg data module."""

    def __init__(
        self: "SmilesImgDataset",
        preds_csv_path: Path = default_path,
        imgs_df_dir: Path = default_path.parents[0],
        batch_size: int = 32,
        transform: Optional[Any] = None,
        target_transform: Optional[Any] = None,
        model: Optional[Union[nn.Module, pl.LightningModule]] = None,
    ) -> None:
        """OSM data module init.

        Args:
            preds_csv_path (Path): csv path.
                Defaults to default_path.
            imgs_df_dir (Path): images csv path.
                Defaults to default_path.parents[0].
            batch_size (int): batch size. Defaults to 32.
            transform (Optional[Any]): Tranforms for the inputs. Defaults to None.
            target_transform (Optional[Any]): Transforms for the target. Defaults to None.
            model (Optional[Union[nn.Module, pl.LightningModule]]): Model. Default to None.
        """
        super().__init__()
        self.preds_csv_path = preds_csv_path
        self.imgs_df_dir = imgs_df_dir
        self.batch_size = batch_size
        self.transform = transform
        self.target_transform = target_transform
        self.model = model

    def prepare_data(self: "SmilesImgDataModule") -> None:
        """Prepare data."""
        # download, split, etc...
        # only called on 1 GPU/TPU in distributed
        pass

    def setup(self: "SmilesImgDataModule", stage: Optional[str]) -> None:
        """Setup dataloaders.

        Args:
            stage (Optional[str]): Optional pipeline state
        """
        self.csv_full = SmilesImgDataset(
            self.preds_csv_path,
            self.imgs_df_dir,
            self.transform,
            self.target_transform,
        )

        train_len = int(len(self.csv_full) * 0.8)
        validate_len = int(len(self.csv_full) * 0.1)
        test_len = int(len(self.csv_full) * 0.1)
        others_len = len(self.csv_full) - train_len - validate_len - test_len
        self.csv_train, self.csv_val, self.csv_test, _ = random_split(
            self.csv_full,
            [train_len, validate_len, test_len, others_len],
        )

        # hack for now, please fixme later
        self.enc_csv_train = self.csv_train
        self.enc_csv_test = self.csv_test
        self.csv_predict = self.csv_test
        self.enc_csv_predict = self.csv_test

    def train_dataloader(self: "SmilesImgDataModule") -> DataLoader:
        """Train dataloader.

        Returns:
            DataLoader: train dataloader
        """
        return DataLoader(
            self.csv_train, batch_size=self.batch_size, num_workers=12, pin_memory=True
        )

    def enc_train_dataloader(
        self: "SmilesImgDataModule", context: ts.Context
    ) -> DataLoader:
        """Encrypted train dataloader.

        Args:
            context (ts.Context): Tenseal encryption context.

        Returns:
            DataLoader: train dataloader
        """
        enc_csv_train = EncConvDataset(
            context,
            self.enc_csv_train,
            self.model.conv1.kernel_size,
            self.model.conv1.stride[0],
            (32, 32),
        )
        return DataLoader(enc_csv_train, batch_size=None)

    def val_dataloader(self: "SmilesImgDataModule") -> DataLoader:
        """Val dataloader.

        Returns:
            DataLoader: val dataloader
        """
        return DataLoader(
            self.csv_val, batch_size=self.batch_size, num_workers=12, pin_memory=True
        )

    def enc_val_dataloader(
        self: "SmilesImgDataModule", context: ts.Context
    ) -> DataLoader:
        """Encrypted val dataloader.

        Args:
            context (ts.Context): Tenseal encryption context.

        Returns:
            DataLoader: val dataloader
        """
        enc_csv_val = EncConvDataset(
            context,
            self.enc_csv_val,
            self.model.conv1.kernel_size,
            self.model.conv1.stride[0],
            (32, 32),
        )
        return DataLoader(enc_csv_val, batch_size=None)

    def test_dataloader(self: "SmilesImgDataModule") -> DataLoader:
        """Test data loader.

        Returns:
            DataLoader : test dataloader
        """
        return DataLoader(self.csv_test, batch_size=self.batch_size)

    def enc_test_dataloader(
        self: "SmilesImgDataModule", context: ts.Context
    ) -> DataLoader:
        """Encrypted test dataloader.

        Args:
            context (ts.Context): Tenseal encryption context.

        Returns:
            DataLoader: test dataloader
        """
        enc_csv_test = EncConvDataset(
            context,
            self.enc_csv_test,
            self.model.conv1.kernel_size,
            self.model.conv1.stride[0],
            (32, 32),
        )
        return DataLoader(enc_csv_test, batch_size=None)

    def predict_dataloader(self: "SmilesImgDataModule") -> DataLoader:
        """Predict data loader.

        Returns:
            DataLoader : predict dataloader
        """
        return DataLoader(self.csv_predict, batch_size=self.batch_size)

    def enc_predict_dataloader(
        self: "SmilesImgDataModule", context: ts.Context
    ) -> DataLoader:
        """Encrypted predict dataloader.

        Args:
            context (ts.Context): Tenseal encryption context.

        Returns:
            DataLoader: predict dataloader
        """
        enc_csv_predict = EncConvDataset(
            context,
            self.enc_csv_predict,
            self.model.conv1.kernel_size,
            self.model.conv1.stride[0],
            (32, 32),
        )
        return DataLoader(enc_csv_predict, batch_size=None)

    def teardown(self: "SmilesImgDataModule", stage: Optional[str] = None) -> None:
        """Teardown of data module."""
        pass



"""Models."""

from .fhe_activation import sigmoid_polyval, softplus_polyval  # noqa: F401
from .fhe_wrapper import FHEOlindaNet, FHEOlindaNetOne, FHEOlindaNetZero  # noqa: F401
from .olinda_net import OlindaNet, OlindaNetOne, OlindaNetZero  # noqa: F401



"""Olinda Regression model training pipeline."""

from pathlib import Path

from kedro.pipeline import node, Pipeline
import pytorch_lightning as pl

from chemxor.data import OlindaRDataModule
from chemxor.model import OlindaNetOne
from chemxor.utils import get_project_root_path


def init_data_module() -> pl.LightningDataModule:
    """Initialize data module."""
    return OlindaRDataModule()


def init_model() -> pl.LightningModule:
    """Initialize model."""
    return OlindaNetOne(output=1)


default_path = get_project_root_path().joinpath(
    "data/06_models/olindanet-one-regression"
)


def train_model(
    data_module: pl.LightningDataModule,
    model: pl.LightningDataModule,
    checkpoint_path: Path = default_path,
) -> None:
    """Train model."""
    checkpoint_callback = pl.callbacks.ModelCheckpoint(
        dirpath=checkpoint_path,
        save_top_k=3,
        monitor="VAL_Loss",
    )
    trainer = pl.Trainer(
        callbacks=[checkpoint_callback],
        accelerator="auto",
        gradient_clip_val=0.5,
        val_check_interval=0.10,
    )
    trainer.fit(model=model, datamodule=data_module)


init_data_module_node = node(func=init_data_module, inputs=None, outputs="data_module")
init_model_node = node(func=init_model, inputs=None, outputs="model")
train_model_node = node(
    func=train_model,
    inputs=["data_module", "model"],
    outputs=None,
)

# Assemble nodes into a pipeline
olinda_one_reg_pipeline = Pipeline(
    [
        init_data_module_node,
        init_model_node,
        train_model_node,
    ]
)



"""FHE compatible activations."""

from typing import Any

import tenseal as ts


def sigmoid_polyval(x: Any) -> Any:
    """Sigmoid polynomial approximation.

    Args:
        x (Any): model input

    Returns:
        Any: model output
    """
    if type(x) in [ts.CKKSTensor, ts.CKKSVector]:
        return x.polyval([0.5, 0.197, 0, -0.004])
    else:
        return 0.5 + (0.199 * x) - (0.004 * (x**3))


def softplus_polyval(x: Any) -> Any:
    """Softplus polynomial approximation.

    Args:
        x (Any): model input

    Returns:
        Any: model output
    """
    if type(x) in [ts.CKKSTensor, ts.CKKSVector]:
        return x.polyval([1.8697581, 0.5, 0.013394318])
    else:
        return 1.8697581 + (0.5 * x) + (0.013394318 * (x**2)) - (-0.000002 * (x**4))



"""Utilities."""

from pathlib import Path
from typing import Any, List, Tuple, Union

import joblib
import numpy as np
from rdkit import Chem
from rdkit.Chem import AllChem
import tenseal as ts
import torch as t

from chemxor.schema.fhe_model import PreProcessInput


def get_project_root_path() -> Path:
    """Get path of the project root.

    Returns:
        Path: Project root path
    """
    return Path(__file__).parents[2].absolute()


def get_package_root_path() -> Path:
    """Get path of the package root.

    Returns:
        Path: Package root path
    """
    return Path(__file__).parents[0].absolute()


def prepare_fhe_input(
    model_output: List,
    pre_processors: List[Tuple[PreProcessInput, List]],
    context: ts.Context,
) -> Union[ts.CKKSTensor, ts.CKKSVector]:
    """Prepare input for next step.

    Args:
        model_output (List): Decrypted model ouput to prepare as an input for next step.
        pre_processors (List[Tuple[PreProcessInput, List]]): List of Input pre processor to use
        context (ts.Context): Tenseal encryption context

    Raises:
        NotImplementedError: Pre processor is not implemented

    Returns:
        Union[ts.CKKSTensor, ts.CKKSVector]: Prepared input
    """
    model_output
    for pre_processor in pre_processors:
        if pre_processor[0] == PreProcessInput.IM_TO_COL:
            image_list = []
            for channel in model_output:
                out, _ = ts.im2col_encoding(context, channel, *pre_processor[1])
                image_list.append(out)
            model_output = image_list
        elif pre_processor[0] == PreProcessInput.RE_ENCRYPT:
            model_output = ts.ckks_vector(context, model_output)
        elif pre_processor[0] == PreProcessInput.RESHAPE:
            model_output = t.tensor(model_output).view(pre_processor[1][0]).tolist()
        elif pre_processor[0] == PreProcessInput.PASSTHROUGH:
            model_output = model_output
        else:
            raise NotImplementedError
    return model_output


def evaluate_fhe_model(model: Any, enc_sample: Any, decrypt: bool = False) -> List:
    """Helper to evaluate model on FHE inputs.

    Args:
        model (Any): model wrapped with FHE wrapper
        enc_sample (Any): FHE encrypted sample
        decrypt (bool): decrypt output. Defaults to False.

    Returns:
        List: Model output
    """
    output = enc_sample
    for step in range(model.steps + 1):
        output = model(output, step)
        dec_out = output.decrypt()
        output = prepare_fhe_input(dec_out, model.pre_process[step], model.enc_context)

    if decrypt is True:
        output = output.decrypt()
    return output


def smiles_to_imcol(x: str, context: ts.Context) -> Any:
    """Convert smiles string to compatible inputs for FHE models.

    Args:
        x (str): SMILES string
        context (ts.Context): TenSeal Context

    Returns:
        Any: Transformed encrypted input
    """
    # Load grid transformer
    project_root_path = get_project_root_path()
    grid_transformer = joblib.load(
        project_root_path.joinpath("data/06_models/grid_transformer.joblib")
    )
    # Create molecule figerprints
    molecule_fp = AllChem.GetMorganFingerprint(
        Chem.MolFromSmiles(x),
        radius=3,
        useCounts=True,
        useFeatures=True,
    )

    # Truncate fp to len 1024
    molecule_fp_truncated = np.zeros((1024), np.uint8)
    for idx, v in molecule_fp.GetNonzeroElements().items():
        nidx = idx % 1024
        molecule_fp_truncated[nidx] += int(v)

    # Create input image
    molecule_img = grid_transformer.transform([molecule_fp_truncated])[0]
    molecule_img = t.tensor(molecule_img, dtype=t.float).reshape(1, 32, 32)

    # Encrypt items
    enc_x, _ = ts.im2col_encoding(
        context,
        molecule_img.view(32, 32).tolist(),
        3,
        3,
        3,
    )
    return enc_x



"""FHE model wrappers."""

from typing import Any

import pytorch_lightning as pl
import tenseal as ts

from chemxor.model.fhe_activation import softplus_polyval
from chemxor.model.olinda_net import OlindaNet, OlindaNetOne, OlindaNetZero
from chemxor.schema.fhe_model import PreProcessInput


class FHEOlindaNetZero(pl.LightningModule):
    """FHEOlindaNet Zero: Wrapper to evaluate FHE inputs."""

    def __init__(self: "FHEOlindaNetZero", model: OlindaNetZero) -> None:
        """Init."""
        super().__init__()

        self._model = model

        # Prepare layers
        self.conv1_weight = model.conv1.weight.data.view(
            model.conv1.in_channels,
            model.conv1.out_channels,
            model.conv1.kernel_size[0],
            model.conv1.kernel_size[1],
        ).tolist()
        self.conv1_bias = model.conv1.bias.data.tolist()

        self.fc1_weight = model.fc1.weight.T.data.tolist()
        self.fc1_bias = model.fc1.bias.data.tolist()

        self.fc2_weight = model.fc2.weight.T.data.tolist()
        self.fc2_bias = model.fc2.bias.data.tolist()

        self.fc3_weight = model.fc3.weight.T.data.tolist()
        self.fc3_bias = model.fc3.bias.data.tolist()

        # Prepare parameters
        self.steps = 2
        self.conv1_windows_nb = 100
        self.pre_process = [
            [(PreProcessInput.RESHAPE, [3200]), (PreProcessInput.RE_ENCRYPT, [])],
            [(PreProcessInput.RE_ENCRYPT, [])],
            [(PreProcessInput.RE_ENCRYPT, [])],
        ]

        # Encryption context
        self.bits_scale = 26
        self.poly_modulus_degree = 8192
        self.coeff_mod_bit_sizes = [
            31,
            self.bits_scale,
            self.bits_scale,
            self.bits_scale,
            self.bits_scale,
            self.bits_scale,
            self.bits_scale,
            31,
        ]
        self.enc_context = ts.context(
            ts.SCHEME_TYPE.CKKS,
            poly_modulus_degree=self.poly_modulus_degree,
            coeff_mod_bit_sizes=self.coeff_mod_bit_sizes,
        )
        self.enc_context.global_scale = pow(2, self.bits_scale)
        self.enc_context.generate_galois_keys()

    def forward(self: "FHEOlindaNetZero", x: Any, step: int) -> Any:
        """Forward function.

        Args:
            x (Any): model input
            step (int): model step

        Returns:
            Any: model output
        """
        if step == 0:
            # conv layer 1
            running_sum = None
            for i, _ in enumerate(self.conv1_weight):
                enc_channels = []
                for kernel, bias in zip(self.conv1_weight[i], self.conv1_bias):
                    y = x.conv2d_im2col(kernel, self.conv1_windows_nb) + bias
                    enc_channels.append(y)
                # pack all channels into a single flattened vector
                enc_x = ts.CKKSVector.pack_vectors(enc_channels)
                if running_sum is None:
                    running_sum = enc_x
                else:
                    running_sum = running_sum + enc_x
            enc_x = softplus_polyval(enc_x)
            return enc_x
        elif step == 1:
            # fc1 layer
            enc_x = x.mm(self.fc1_weight) + self.fc1_bias
            enc_x = softplus_polyval(enc_x)
            # fc2 layer
            enc_x = enc_x.mm(self.fc2_weight) + self.fc2_bias
            enc_x = softplus_polyval(enc_x)
            return enc_x
        elif step == 2:
            # fc3 layer
            enc_x = x.mm(self.fc3_weight) + self.fc3_bias
            return enc_x


class FHEOlindaNet(pl.LightningModule):
    """FHEOlindaNet: Wrapper to evaluate FHE inputs."""

    def __init__(self: "FHEOlindaNet", model: OlindaNet) -> None:
        """Init."""
        super().__init__()

        self._model = model

        # Prepare layers
        self.conv1_weight = model.conv1.weight.data.view(
            model.conv1.in_channels,
            model.conv1.out_channels,
            model.conv1.kernel_size[0],
            model.conv1.kernel_size[1],
        ).tolist()
        self.conv1_bias = model.conv1.bias.data.tolist()

        self.conv2_weight = model.conv2.weight.data.view(
            model.conv2.in_channels,
            model.conv2.out_channels,
            model.conv2.kernel_size[0],
            model.conv2.kernel_size[1],
        ).tolist()
        self.conv2_bias = model.conv2.bias.data.tolist()

        self.fc1_weight = model.fc1.weight.T.data.tolist()
        self.fc1_bias = model.fc1.bias.data.tolist()

        self.fc2_weight = model.fc2.weight.T.data.tolist()
        self.fc2_bias = model.fc2.bias.data.tolist()

        self.fc3_weight = model.fc3.weight.T.data.tolist()
        self.fc3_bias = model.fc3.bias.data.tolist()

        self.fc4_weight = model.fc4.weight.T.data.tolist()
        self.fc4_bias = model.fc4.bias.data.tolist()

        # Prepare parameters
        self.steps = 3
        self.conv1_windows_nb = 100
        self.conv2_windows_nb = 64

        self.pre_process = [
            [
                (PreProcessInput.RESHAPE, [(32, 10, 10)]),
                (PreProcessInput.IM_TO_COL, [3, 3, 1]),
            ],
            [(PreProcessInput.RESHAPE, [2048]), (PreProcessInput.RE_ENCRYPT, [])],
            [(PreProcessInput.RE_ENCRYPT, [])],
            [(PreProcessInput.RE_ENCRYPT, [])],
        ]

        # Encryption context
        self.bits_scale = 26
        self.poly_modulus_degree = 8192
        self.coeff_mod_bit_sizes = [
            31,
            self.bits_scale,
            self.bits_scale,
            self.bits_scale,
            self.bits_scale,
            self.bits_scale,
            self.bits_scale,
            31,
        ]
        self.enc_context = ts.context(
            ts.SCHEME_TYPE.CKKS,
            poly_modulus_degree=self.poly_modulus_degree,
            coeff_mod_bit_sizes=self.coeff_mod_bit_sizes,
        )
        self.enc_context.global_scale = pow(2, self.bits_scale)
        self.enc_context.generate_galois_keys()

    def forward(self: "FHEOlindaNet", x: Any, step: int) -> Any:
        """Forward function.

        Args:
            x (Any): model input
            step (int): model step

        Returns:
            Any: model output
        """
        if step == 0:
            # conv layer 1
            running_sum = None
            for i, _ in enumerate(self.conv1_weight):
                enc_channels = []
                for kernel, bias in zip(self.conv1_weight[i], self.conv1_bias):
                    y = x.conv2d_im2col(kernel, self.conv1_windows_nb) + bias
                    enc_channels.append(y)
                # pack all channels into a single flattened vector
                enc_x = ts.CKKSVector.pack_vectors(enc_channels)
                if running_sum is None:
                    running_sum = enc_x
                else:
                    running_sum = running_sum + enc_x
            enc_x = softplus_polyval(enc_x)
            return enc_x
        elif step == 1:
            # conv layer 2
            running_sum = None
            for i, _ in enumerate(self.conv2_weight):
                enc_channels = []
                for kernel, bias in zip(self.conv2_weight[i], self.conv2_bias):
                    y = x[i].conv2d_im2col(kernel, self.conv2_windows_nb) + bias
                    enc_channels.append(y)
                # pack all channels into a single flattened vector
                enc_x = ts.CKKSVector.pack_vectors(enc_channels)
                if running_sum is None:
                    running_sum = enc_x
                else:
                    running_sum = running_sum + enc_x
            enc_x = softplus_polyval(enc_x)
            return enc_x
        elif step == 2:
            # fc1 layer
            enc_x = x.mm(self.fc1_weight) + self.fc1_bias
            enc_x = softplus_polyval(enc_x)
            # fc2 layer
            enc_x = enc_x.mm(self.fc2_weight) + self.fc2_bias
            enc_x = softplus_polyval(enc_x)
            return enc_x

        elif step == 3:
            # fc3 layer
            enc_x = x.mm(self.fc3_weight) + self.fc3_bias
            enc_x = softplus_polyval(enc_x)
            # fc4 layer
            enc_x = enc_x.mm(self.fc4_weight) + self.fc4_bias
            return enc_x


class FHEOlindaNetOne(pl.LightningModule):
    """FHEOlindaNet One: Wrapper to evaluate FHE inputs."""

    def __init__(self: "FHEOlindaNetOne", model: OlindaNetOne) -> None:
        """Init."""
        super().__init__()

        self._model = model

        # Prepare layers
        self.conv1_weight = model.conv1.weight.data.view(
            model.conv1.in_channels,
            model.conv1.out_channels,
            model.conv1.kernel_size[0],
            model.conv1.kernel_size[1],
        ).tolist()
        self.conv1_bias = model.conv1.bias.data.tolist()

        self.conv2_weight = model.conv2.weight.data.view(
            model.conv2.in_channels,
            model.conv2.out_channels,
            model.conv2.kernel_size[0],
            model.conv2.kernel_size[1],
        ).tolist()
        self.conv2_bias = model.conv2.bias.data.tolist()

        self.conv3_weight = model.conv3.weight.data.view(
            model.conv3.in_channels,
            model.conv3.out_channels,
            model.conv3.kernel_size[0],
            model.conv3.kernel_size[1],
        ).tolist()
        self.conv3_bias = model.conv3.bias.data.tolist()

        self.conv4_weight = model.conv4.weight.data.view(
            model.conv4.in_channels,
            model.conv4.out_channels,
            model.conv4.kernel_size[0],
            model.conv4.kernel_size[1],
        ).tolist()
        self.conv4_bias = model.conv4.bias.data.tolist()

        self.fc1_weight = model.fc1.weight.T.data.tolist()
        self.fc1_bias = model.fc1.bias.data.tolist()

        self.fc2_weight = model.fc2.weight.T.data.tolist()
        self.fc2_bias = model.fc2.bias.data.tolist()

        self.fc3_weight = model.fc3.weight.T.data.tolist()
        self.fc3_bias = model.fc3.bias.data.tolist()

        self.fc4_weight = model.fc4.weight.T.data.tolist()
        self.fc4_bias = model.fc4.bias.data.tolist()

        # Prepare parameters
        self.steps = 5
        self.conv1_windows_nb = 100
        self.conv2_windows_nb = 64
        self.conv3_windows_nb = 36
        self.conv4_windows_nb = 16

        self.pre_process = [
            [
                (PreProcessInput.RESHAPE, [(32, 10, 10)]),
                (PreProcessInput.IM_TO_COL, [3, 3, 1]),
            ],
            [
                (PreProcessInput.RESHAPE, [(32, 8, 8)]),
                (PreProcessInput.IM_TO_COL, [3, 3, 1]),
            ],
            [
                (PreProcessInput.RESHAPE, [(64, 6, 6)]),
                (PreProcessInput.IM_TO_COL, [3, 3, 1]),
            ],
            [(PreProcessInput.RESHAPE, [(1024)]), (PreProcessInput.RE_ENCRYPT, [])],
            [(PreProcessInput.RE_ENCRYPT, [])],
            [(PreProcessInput.RE_ENCRYPT, [])],
        ]

        # Encryption context
        self.bits_scale = 26
        self.poly_modulus_degree = 8192
        self.coeff_mod_bit_sizes = [
            31,
            self.bits_scale,
            self.bits_scale,
            self.bits_scale,
            self.bits_scale,
            self.bits_scale,
            self.bits_scale,
            31,
        ]
        self.enc_context = ts.context(
            ts.SCHEME_TYPE.CKKS,
            poly_modulus_degree=self.poly_modulus_degree,
            coeff_mod_bit_sizes=self.coeff_mod_bit_sizes,
        )
        self.enc_context.global_scale = pow(2, self.bits_scale)
        self.enc_context.generate_galois_keys()

    def forward(self: "FHEOlindaNetOne", x: Any, step: int) -> Any:
        """Forward function.

        Args:
            x (Any): model input
            step (int): model step

        Returns:
            Any: model output
        """
        if step == 0:
            # conv layer 1
            running_sum = None
            for i, _ in enumerate(self.conv1_weight):
                enc_channels = []
                for kernel, bias in zip(self.conv1_weight[i], self.conv1_bias):
                    y = x.conv2d_im2col(kernel, self.conv1_windows_nb) + bias
                    enc_channels.append(y)
                # pack all channels into a single flattened vector
                enc_x = ts.CKKSVector.pack_vectors(enc_channels)
                if running_sum is None:
                    running_sum = enc_x
                else:
                    running_sum = running_sum + enc_x
            enc_x = softplus_polyval(enc_x)
            return enc_x
        elif step == 1:
            # conv layer 2
            running_sum = None
            for i, _ in enumerate(self.conv2_weight):
                enc_channels = []
                for kernel, bias in zip(self.conv2_weight[i], self.conv2_bias):
                    y = x[i].conv2d_im2col(kernel, self.conv2_windows_nb) + bias
                    enc_channels.append(y)
                # pack all channels into a single flattened vector
                enc_x = ts.CKKSVector.pack_vectors(enc_channels)
                if running_sum is None:
                    running_sum = enc_x
                else:
                    running_sum = running_sum + enc_x
            enc_x = softplus_polyval(enc_x)
            return enc_x
        elif step == 2:
            # conv layer 3
            running_sum = None
            for i, _ in enumerate(self.conv3_weight):
                enc_channels = []
                for kernel, bias in zip(self.conv3_weight[i], self.conv3_bias):
                    y = x[i].conv2d_im2col(kernel, self.conv3_windows_nb) + bias
                    enc_channels.append(y)
                # pack all channels into a single flattened vector
                enc_x = ts.CKKSVector.pack_vectors(enc_channels)
                if running_sum is None:
                    running_sum = enc_x
                else:
                    running_sum = running_sum + enc_x
            enc_x = softplus_polyval(enc_x)
            return enc_x
        elif step == 3:
            # conv layer 4
            running_sum = None
            for i, _ in enumerate(self.conv4_weight):
                enc_channels = []
                for kernel, bias in zip(self.conv4_weight[i], self.conv4_bias):
                    y = x[i].conv2d_im2col(kernel, self.conv4_windows_nb) + bias
                    enc_channels.append(y)
                # pack all channels into a single flattened vector
                enc_x = ts.CKKSVector.pack_vectors(enc_channels)
                if running_sum is None:
                    running_sum = enc_x
                else:
                    running_sum = running_sum + enc_x
            enc_x = softplus_polyval(enc_x)
            return enc_x
        elif step == 4:
            # fc1 layer
            enc_x = x.mm(self.fc1_weight) + self.fc1_bias
            enc_x = softplus_polyval(enc_x)
            # fc2 layer
            enc_x = enc_x.mm(self.fc2_weight) + self.fc2_bias
            enc_x = softplus_polyval(enc_x)
            return enc_x

        elif step == 5:
            # fc3 layer
            enc_x = x.mm(self.fc3_weight) + self.fc3_bias
            enc_x = softplus_polyval(enc_x)
            # fc4 layer
            enc_x = enc_x.mm(self.fc4_weight) + self.fc4_bias
            return enc_x



"""Smiles to Image pre processor pipeline."""

from pathlib import Path
from typing import Any, List

import joblib
from joblib import cpu_count, delayed, Parallel
from kedro.pipeline import node, Pipeline
import numpy as np
import pandas as pd
import pyarrow.feather as feather
from rdkit import Chem
from rdkit.Chem import AllChem
from tqdm import tqdm

from chemxor.utils import get_project_root_path


project_root_path = get_project_root_path()
default_path = project_root_path.joinpath(
    "data/01_raw/ModelPreds/eos2r5a/ersilia_output.csv"
)
transformer_path = project_root_path.joinpath("data/06_models/grid_transformer.joblib")


def chunker(seq: List, size: int) -> List:
    """Create bacthes for processing smiles.

    Args:
        seq (List): Array of smiles
        size (int): batch size

    Returns:
        List: List of batches
    """
    return (seq[pos : pos + size] for pos in range(0, len(seq), size))


def ecfp_counts(mols: List) -> List:
    """Create ECFPs from batch of smiles.

    Args:
        mols (List): batch of smiles

    Returns:
        List: batch of ECFPs
    """
    fps = [
        AllChem.GetMorganFingerprint(mol, radius=3, useCounts=True, useFeatures=True)
        for mol in mols
    ]
    nfp = np.zeros((len(fps), 1024), np.uint8)
    for i, fp in enumerate(fps):
        for idx, v in fp.GetNonzeroElements().items():
            nidx = idx % 1024
            nfp[i, nidx] += int(v)
    return nfp


def convert_smiles_to_imgs(
    in_path: Path,
    out_path: Path,
    transformer_path: Path = transformer_path,
    df_chunks: int = 10000,
) -> None:
    """Convert Smiles to Images.

    Args:
        in_path (Path): Input CSV path
        out_path (Path): Output CSV path
        transformer_path (Path): Path of transformer. Defaults to transformer_path.
        df_chunks (int): df chunks to create. Defaults to 100.
    """
    df_full = pd.read_csv(project_root_path.joinpath(in_path).absolute())
    df_len = len(df_full)
    splits = [x for x in range(0, df_len + 1, (df_len + 1) // df_chunks)]
    split_tuples = [(splits[i], splits[i + 1]) for i in range(len(splits) - 1)]
    dfs = [df_full.iloc[start:end] for start, end in split_tuples]
    grid_transformer = joblib.load(transformer_path)
    Parallel(n_jobs=cpu_count())(
        delayed(chunked_df_to_feather)(df, i, grid_transformer, in_path, split_tuples)
        for i, df in enumerate(tqdm(dfs))
    )


def chunked_df_to_feather(
    df: pd.DataFrame, i: int, grid_transformer: Any, in_path: Path, split_tuples: tuple
) -> None:
    """Save chunked dfs to feather files.

    Args:
        df (pd.DataFrame): pandas Dataframe
        i (int): dataframe position
        grid_transformer (Any): grid transformer
        in_path (Path): Path
        split_tuples (tuple): Split ranges
    """
    smiles = df["smiles"]
    R = []
    for chunk in tqdm(chunker(smiles, 100)):
        mols = [Chem.MolFromSmiles(smi) for smi in chunk]
        e = ecfp_counts(mols)
        R += [e]
    ecfp = np.concatenate(R)
    molecule_imgs = grid_transformer.transform(ecfp)
    imgs_df = pd.DataFrame(molecule_imgs.reshape(-1, 1024))
    global_index_df = pd.DataFrame(
        [
            x
            for x in range(
                split_tuples[i][0],
                split_tuples[i][1],
            )
        ],
        columns=["global_index"],
    )
    global_index_df["target"] = df.iloc[:, 0].to_list()
    global_index_df[[x for x in range(0, 1024)]] = imgs_df
    feather.write_feather(
        global_index_df,
        project_root_path.joinpath(in_path)
        .parents[0]
        .joinpath(f"sm_to_imgs_{i}.feather.lz4")
        .absolute(),
        compression="lz4",
    )


convert_smiles_to_imgs_node = node(
    func=convert_smiles_to_imgs,
    inputs=["params:in_path", "params:out_path"],
    outputs=None,
)

# Assemble nodes into a pipeline
convert_smiles_to_imgs_pipeline = Pipeline([convert_smiles_to_imgs_node])



"""FHE Encrypted model API schema."""

from enum import Enum
from typing import List, Optional, Tuple

from pydantic import BaseModel, Field


class PreProcessInput(Enum):
    """Supported input pre processor."""

    IM_TO_COL = "im_to_col"
    RE_ENCRYPT = "re_encrypt"
    RESHAPE = "reshape"
    PASSTHROUGH = "passthrough"


class FHEModelQueryPostRequest(BaseModel):
    """FHE Encrypted model query post request schema."""

    ts_context: str = Field(unicode_safe=False)
    model_input: str = Field(unicode_safe=False)


class FHEModelQueryPostResponse(BaseModel):
    """FHE Encrypted model query post response schema."""

    model_output: Optional[str] = Field(unicode_safe=False)


class PartFHEModelQueryPostRequest(FHEModelQueryPostRequest):
    """Partitioned FHE Encrypted model query post request schema."""

    model_step: Optional[int]


class PartFHEModelQueryPostResponse(FHEModelQueryPostResponse):
    """Partitioned FHE Encrypted model query post response schema."""

    next_step: Optional[int]
    preprocess_next_args: List[Tuple[PreProcessInput, List]]


class ModelContextParams(BaseModel):
    """Model tenseal context parameters."""

    bit_scale: int
    poly_modulus_degree: int
    coeff_mod_bit_sizes: List[int]


class ModelInfo(BaseModel):
    """Model info container."""

    model_name: str
    model_steps: int
    context_params: ModelContextParams


class PartFHEModelQueryGetResponse(BaseModel):
    """Partitioned FHE Encrypted model query get response schema."""

    model_info: ModelInfo



"""Test encryption schema."""

from pydantic import BaseModel, Field


class TestEncPost(BaseModel):
    """Test encryption post schema."""

    context: str = Field(unicode_safe=False)
    vector: str = Field(unicode_safe=False)


class TestEncResponse(BaseModel):
    """Test encryption response schema."""

    vector: str = Field(unicode_safe=False)



"""OlindaNet modules."""

from typing import Any, Callable, Optional

import pytorch_lightning as pl
from torch import nn
from torch.nn import functional as F
from torch.optim import Adam, Optimizer
from torchmetrics import Accuracy, MeanSquaredError, MetricCollection, Precision, Recall

from chemxor.model.fhe_activation import softplus_polyval


class OlindaNetZero(pl.LightningModule):
    """OlindaNet Zero: Slim(relatively) distillation network."""

    def __init__(
        self: "OlindaNetZero", output: int = 10, criterion: Optional[Callable] = None
    ) -> None:
        """Init."""
        super().__init__()
        self.output = output
        self.conv1 = nn.Conv2d(1, 32, kernel_size=3, padding=0, stride=3)
        self.fc1 = nn.Linear(3200, 512)
        self.fc2 = nn.Linear(512, 256)
        self.fc3 = nn.Linear(256, output)

        # Classification Metrics
        if output > 1:
            metrics = MetricCollection(
                [
                    Accuracy(),
                    Precision(num_classes=output, average="macro"),
                    Recall(num_classes=output, average="macro"),
                ]
            )
        else:
            metrics = MetricCollection([MeanSquaredError()])

        self.train_metrics = metrics.clone(prefix="TRAIN_")
        self.valid_metrics = metrics.clone(prefix="VAL_")
        self.test_metrics = metrics.clone(prefix="TEST_")

        # Criterion
        if criterion is None:
            if output > 1:
                self.criterion = F.cross_entropy
            else:
                self.criterion = F.mse_loss
        else:
            self.criterion = criterion

    def forward(self: "OlindaNetZero", x: Any) -> Any:
        """Forward function.

        Args:
            x (Any): model input

        Returns:
            Any: model output
        """
        x = self.conv1(x)
        x = softplus_polyval(x)

        # flattening while keeping the batch axis
        x = x.view(x.shape[0], -1)
        x = self.fc1(x)
        x = softplus_polyval(x)
        x = self.fc2(x)
        x = softplus_polyval(x)
        x = self.fc3(x)
        return x

    def training_step(self: "OlindaNetZero", batch: Any, batch_idx: Any) -> Any:
        """Training step.

        Args:
            batch (Any): input batch
            batch_idx (Any): batch index

        Returns:
            Any: training step loss
        """
        x, y = batch
        output = self(x)
        loss = self.criterion(output, y)

        # Logging to TensorBoard by default
        self.log("TRAIN_Loss", loss)

        # Logging metrics
        if self.output > 1:
            y = y.int()
        metrics = self.train_metrics(output, y)
        self.log_dict(metrics)
        return loss

    def validation_step(self: "OlindaNetZero", batch: Any, batch_idx: Any) -> None:
        """Validation step.

        Args:
            batch (Any): input batch
            batch_idx (Any): batch index
        """
        x, y = batch
        output = self(x)
        loss = self.criterion(output, y)

        # Logging to TensorBoard by default
        self.log("VAL_Loss", loss)

        # Logging metrics
        if self.output > 1:
            y = y.int()
        metrics = self.valid_metrics(output, y)
        self.log_dict(metrics)

    def test_step(self: "OlindaNetZero", batch: Any, batch_idx: Any) -> None:
        """Test step.

        Args:
            batch (Any): input batch
            batch_idx (Any): batch index
        """
        x, y = batch
        output = self(x)
        loss = self.criterion(output, y)

        # Logging to TensorBoard by default
        self.log("TEST_Loss", loss)

        # Logging metrics
        if self.output > 1:
            y = y.int()
        metrics = self.test_metrics(output, y)
        self.log_dict(metrics)

    def configure_optimizers(self: "OlindaNetZero") -> Optimizer:
        """Configure optimizer."""
        optimizer = Adam(self.parameters(), lr=0.001)
        return optimizer


class OlindaNet(pl.LightningModule):
    """OlindaNet: A good compromise(relatively) distillation network."""

    def __init__(
        self: "OlindaNet", output: int = 10, criterion: Optional[Callable] = None
    ) -> None:
        """Init."""
        super().__init__()
        self.output = output
        self.conv1 = nn.Conv2d(1, 32, kernel_size=3, padding=0, stride=3)
        self.conv2 = nn.Conv2d(32, 32, kernel_size=3, padding=0, stride=1)
        self.fc1 = nn.Linear(2048, 1024)
        self.fc2 = nn.Linear(1024, 512)
        self.fc3 = nn.Linear(512, 256)
        self.fc4 = nn.Linear(256, output)

        # Classification Metrics
        if output > 1:
            metrics = MetricCollection(
                [
                    Accuracy(),
                    Precision(num_classes=output, average="macro"),
                    Recall(num_classes=output, average="macro"),
                ]
            )
        else:
            metrics = MetricCollection([MeanSquaredError()])

        self.train_metrics = metrics.clone(prefix="TRAIN_")
        self.valid_metrics = metrics.clone(prefix="VAL_")
        self.test_metrics = metrics.clone(prefix="TEST_")

        # Criterion
        if criterion is None:
            if output > 1:
                self.criterion = F.cross_entropy
            else:
                self.criterion = F.mse_loss
        else:
            self.criterion = criterion

    def forward(self: "OlindaNet", x: Any) -> Any:
        """Forward function.

        Args:
            x (Any): model input

        Returns:
            Any: model output
        """
        x = self.conv1(x)
        x = softplus_polyval(x)
        x = self.conv2(x)
        x = softplus_polyval(x)

        # flattening while keeping the batch axis
        x = x.view(x.shape[0], -1)
        x = self.fc1(x)
        x = softplus_polyval(x)
        x = self.fc2(x)
        x = softplus_polyval(x)
        x = self.fc3(x)
        x = softplus_polyval(x)
        x = self.fc4(x)
        return x

    def training_step(self: "OlindaNet", batch: Any, batch_idx: Any) -> Any:
        """Training step.

        Args:
            batch (Any): input batch
            batch_idx (Any): batch index

        Returns:
            Any: training step loss
        """
        x, y = batch
        output = self(x)
        loss = self.criterion(output, y)

        # Logging to TensorBoard by default
        self.log("TRAIN_Loss", loss)

        # Logging metrics
        if self.output > 1:
            y = y.int()
        metrics = self.train_metrics(output, y)
        self.log_dict(metrics)
        return loss

    def validation_step(self: "OlindaNet", batch: Any, batch_idx: Any) -> None:
        """Validation step.

        Args:
            batch (Any): input batch
            batch_idx (Any): batch index
        """
        x, y = batch
        output = self(x)
        loss = self.criterion(output, y)

        # Logging to TensorBoard by default
        self.log("VAL_Loss", loss)

        # Logging metrics
        if self.output > 1:
            y = y.int()
        metrics = self.valid_metrics(output, y)
        self.log_dict(metrics)

    def test_step(self: "OlindaNet", batch: Any, batch_idx: Any) -> None:
        """Test step.

        Args:
            batch (Any): input batch
            batch_idx (Any): batch index
        """
        x, y = batch
        output = self(x)
        loss = self.criterion(output, y)

        # Logging to TensorBoard by default
        self.log("TEST_Loss", loss)

        # Logging metrics
        if self.output > 1:
            y = y.int()
        metrics = self.test_metrics(output, y)
        self.log_dict(metrics)

    def configure_optimizers(self: "OlindaNet") -> Optimizer:
        """Configure optimizer."""
        optimizer = Adam(self.parameters(), lr=0.001)
        return optimizer


class OlindaNetOne(pl.LightningModule):
    """OlindaNet One: Heavy(relatively) distillation network."""

    def __init__(
        self: "OlindaNetOne", output: int = 10, criterion: Optional[Callable] = None
    ) -> None:
        """Init."""
        super().__init__()
        self.output = output
        self.conv1 = nn.Conv2d(1, 32, kernel_size=3, padding=0, stride=3)
        self.conv2 = nn.Conv2d(32, 32, kernel_size=3, padding=0, stride=1)
        self.conv3 = nn.Conv2d(32, 64, kernel_size=3, padding=0, stride=1)
        self.conv4 = nn.Conv2d(64, 64, kernel_size=3, padding=0, stride=1)
        self.fc1 = nn.Linear(1024, 1024)
        self.fc2 = nn.Linear(1024, 512)
        self.fc3 = nn.Linear(512, 256)
        self.fc4 = nn.Linear(256, output)

        # Classification Metrics
        if output > 1:
            metrics = MetricCollection(
                [
                    Accuracy(),
                    Precision(num_classes=output, average="macro"),
                    Recall(num_classes=output, average="macro"),
                ]
            )
        else:
            metrics = MetricCollection([MeanSquaredError()])

        self.train_metrics = metrics.clone(prefix="TRAIN_")
        self.valid_metrics = metrics.clone(prefix="VAL_")
        self.test_metrics = metrics.clone(prefix="TEST_")

        # Criterion
        if criterion is None:
            if output > 1:
                self.criterion = F.cross_entropy
            else:
                self.criterion = F.mse_loss
        else:
            self.criterion = criterion

    def forward(self: "OlindaNetOne", x: Any) -> Any:
        """Forward function.

        Args:
            x (Any): model input

        Returns:
            Any: model output
        """
        x = self.conv1(x)
        x = softplus_polyval(x)
        x = self.conv2(x)
        x = softplus_polyval(x)
        x = self.conv3(x)
        x = softplus_polyval(x)
        x = self.conv4(x)
        x = softplus_polyval(x)

        # flattening while keeping the batch axis
        x = x.view(x.shape[0], -1)
        x = self.fc1(x)
        x = softplus_polyval(x)
        x = self.fc2(x)
        x = softplus_polyval(x)
        x = self.fc3(x)
        x = softplus_polyval(x)
        x = self.fc4(x)
        return x

    def training_step(self: "OlindaNetOne", batch: Any, batch_idx: Any) -> Any:
        """Training step.

        Args:
            batch (Any): input batch
            batch_idx (Any): batch index

        Returns:
            Any: training step loss
        """
        x, y = batch
        output = self(x)
        loss = self.criterion(output, y)

        # Logging to TensorBoard by default
        self.log("TRAIN_Loss", loss)

        # Logging metrics
        if self.output > 1:
            y = y.int()
        metrics = self.train_metrics(output, y)
        self.log_dict(metrics)
        return loss

    def validation_step(self: "OlindaNetOne", batch: Any, batch_idx: Any) -> None:
        """Validation step.

        Args:
            batch (Any): input batch
            batch_idx (Any): batch index
        """
        x, y = batch
        output = self(x)
        loss = self.criterion(output, y)

        # Logging to TensorBoard by default
        self.log("VAL_Loss", loss)

        # Logging metrics
        if self.output > 1:
            y = y.int()
        metrics = self.valid_metrics(output, y)
        self.log_dict(metrics)

    def test_step(self: "OlindaNetOne", batch: Any, batch_idx: Any) -> None:
        """Test step.

        Args:
            batch (Any): input batch
            batch_idx (Any): batch index
        """
        x, y = batch
        output = self(x)
        loss = self.criterion(output, y)

        # Logging to TensorBoard by default
        self.log("TEST_Loss", loss)

        # Logging metrics
        if self.output > 1:
            y = y.int()
        metrics = self.test_metrics(output, y)
        self.log_dict(metrics)

    def configure_optimizers(self: "OlindaNetOne") -> Optimizer:
        """Configure optimizer."""
        optimizer = Adam(self.parameters(), lr=0.001)
        return optimizer



"""Olinda classification data module."""

from copy import deepcopy
from pathlib import Path
from typing import Any, List, Optional, Union

import joblib
import numpy as np
import pandas as pd
import pytorch_lightning as pl
from rdkit import Chem
from rdkit.Chem import AllChem
import tenseal as ts
import torch as t
from torch.utils.data import DataLoader, Dataset, random_split

from chemxor.data.enc_conv_dataset import EncConvDataset
from chemxor.model import OlindaNet, OlindaNetOne, OlindaNetZero
from chemxor.utils import get_project_root_path

project_root_path = get_project_root_path()
default_path = project_root_path.joinpath(
    "data/01_raw/ModelPreds/eos2r5a/ersilia_output.csv"
)


class OlindaCDataset(Dataset):
    """Olinda Classification Dataset."""

    def __init__(
        self: "OlindaCDataset",
        csv_path: Path = default_path,
        threshold: List[float] = [0.5],
        transform: Optional[Any] = None,
        target_transform: Optional[Any] = None,
    ) -> None:
        """Olinda Regression Dataset.

        Args:
            csv_path (Path): csv path.
                Defaults to default_path.
            threshold: List[float]: Threshold values to create categories.
            transform (Optional[Any]): Tranforms for the inputs. Defaults to None.
            target_transform (Optional[Any]): Transforms for the target. Defaults to None.

        """
        self.df_full = pd.read_csv(csv_path.absolute())
        self.threshold = threshold
        self.transform = transform
        self.target_transform = target_transform
        # load the transformer
        self.grid_transformer = joblib.load(
            project_root_path.joinpath("data/06_models/grid_transformer.joblib")
        )

    def __len__(self: "OlindaCDataset") -> int:
        """Length of the dataset.

        Returns:
            int: Length of the dataset.
        """
        return len(self.df_full)

    def __getitem__(self: "OlindaCDataset", index: int) -> Any:
        """Get item from the dataset.

        Args:
            index (int): index of the item.

        Returns:
            Any: Item
        """
        # Extract smile string
        molecule_smile = self.df_full.iloc[index][1]

        # Extract label
        label = float(self.df_full.iloc[index][0])

        # One-hot encoding
        one_hot_tensor = t.zeros((len(self.threshold) + 1))
        for i, value in enumerate(self.threshold):
            if label <= value:
                one_hot_tensor[i] = 1
        if sum(one_hot_tensor).item() == 0:
            one_hot_tensor[-1] = 1

        # Create molecule figerprints
        molecule_fp = AllChem.GetMorganFingerprint(
            Chem.MolFromSmiles(molecule_smile),
            radius=3,
            useCounts=True,
            useFeatures=True,
        )

        # Truncate fp to len 1024
        molecule_fp_truncated = np.zeros((1024), np.uint8)
        for idx, v in molecule_fp.GetNonzeroElements().items():
            nidx = idx % 1024
            molecule_fp_truncated[nidx] += int(v)

        # Create input image
        molecule_img = self.grid_transformer.transform([molecule_fp_truncated])[0]
        molecule_img = t.tensor(molecule_img, dtype=t.float).reshape(1, 32, 32)

        # Apply input transforms
        if self.transform:
            molecule_img = self.transform(molecule_img)

        # Apply target transforms
        if self.target_transform:
            label = self.target_transform(label)

        # Return input and target
        return molecule_img, one_hot_tensor


class OlindaCDataModule(pl.LightningDataModule):
    """Olinda Classification data module."""

    def __init__(
        self: "OlindaCDataset",
        csv_path: Path = default_path,
        threshold: List[float] = [0.5],
        batch_size: int = 32,
        transform: Optional[Any] = None,
        target_transform: Optional[Any] = None,
        model: Optional[Union[OlindaNet, OlindaNetOne, OlindaNetZero]] = None,
    ) -> None:
        """Olinda Regression data module init.

        Args:
            csv_path (Path): csv path.
                Defaults to default_path.
            threshold: List[float]: Threshold values to create categories.
            batch_size (int): batch size. Defaults to 32.
            transform (Optional[Any]): Tranforms for the inputs. Defaults to None.
            target_transform (Optional[Any]): Transforms for the target. Defaults to None.
            model (Optional[Union[OlindaNet, OlindaNetOne, OlindaNetZero]]): Model. Default to None.
        """
        super().__init__()
        self.csv_path = csv_path
        self.threshold = threshold
        self.batch_size = batch_size
        self.transform = transform
        self.target_transform = target_transform
        self.model = model

    def prepare_data(self: "OlindaCDataModule") -> None:
        """Prepare data."""
        # download, split, etc...
        # only called on 1 GPU/TPU in distributed
        pass

    def setup(self: "OlindaCDataModule", stage: Optional[str]) -> None:
        """Setup dataloaders.

        Args:
            stage (Optional[str]): Optional pipeline state
        """
        self.dataset = OlindaCDataset(
            self.csv_path,
            self.threshold,
            self.transform,
            self.target_transform,
        )

        # Calculate split
        train_len = int(len(self.dataset) * 0.7)
        validate_len = int(len(self.dataset) * 0.2)
        test_len = int(len(self.dataset) * 0.1)
        others_len = len(self.dataset) - train_len - validate_len - test_len

        self.csv_train, self.csv_val, self.csv_test, _ = random_split(
            self.dataset,
            [train_len, validate_len, test_len, others_len],
            t.Generator().manual_seed(7777),
        )

        # hack for now, please fixme later
        self.enc_csv_train = deepcopy(self.csv_train)
        self.enc_csv_test = deepcopy(self.csv_test)
        self.csv_predict = deepcopy(self.csv_test)
        self.enc_csv_predict = deepcopy(self.csv_test)

    def train_dataloader(self: "OlindaCDataModule") -> DataLoader:
        """Train dataloader.

        Returns:
            DataLoader: train dataloader
        """
        return DataLoader(self.csv_train, batch_size=self.batch_size)

    def enc_train_dataloader(
        self: "OlindaCDataModule", context: ts.Context
    ) -> DataLoader:
        """Encrypted train dataloader.

        Args:
            context (ts.Context): Tenseal encryption context.

        Returns:
            DataLoader: train dataloader
        """
        enc_csv_train = EncConvDataset(
            context,
            self.enc_csv_train,
            self.model._model.conv1.kernel_size,
            self.model._model.conv1.stride[0],
            (32, 32),
        )
        return DataLoader(enc_csv_train, batch_size=None)

    def val_dataloader(self: "OlindaCDataModule") -> DataLoader:
        """Val dataloader.

        Returns:
            DataLoader: val dataloader
        """
        return DataLoader(self.csv_val, batch_size=self.batch_size)

    def enc_val_dataloader(
        self: "OlindaCDataModule", context: ts.Context
    ) -> DataLoader:
        """Encrypted val dataloader.

        Args:
            context (ts.Context): Tenseal encryption context.

        Returns:
            DataLoader: val dataloader
        """
        enc_csv_val = EncConvDataset(
            context,
            self.enc_csv_val,
            self.model._model.conv1.kernel_size,
            self.model._model.conv1.stride[0],
            (32, 32),
        )
        return DataLoader(enc_csv_val, batch_size=None)

    def test_dataloader(self: "OlindaCDataModule") -> DataLoader:
        """Test data loader.

        Returns:
            DataLoader : test dataloader
        """
        return DataLoader(self.csv_test, batch_size=self.batch_size)

    def enc_test_dataloader(
        self: "OlindaCDataModule", context: ts.Context
    ) -> DataLoader:
        """Encrypted test dataloader.

        Args:
            context (ts.Context): Tenseal encryption context.

        Returns:
            DataLoader: test dataloader
        """
        enc_csv_test = EncConvDataset(
            context,
            self.enc_csv_test,
            self.model._model.conv1.kernel_size,
            self.model._model.conv1.stride[0],
            (32, 32),
        )
        return DataLoader(enc_csv_test, batch_size=None)

    def predict_dataloader(self: "OlindaCDataModule") -> DataLoader:
        """Predict data loader.

        Returns:
            DataLoader : predict dataloader
        """
        return DataLoader(self.csv_predict, batch_size=self.batch_size)

    def enc_predict_dataloader(
        self: "OlindaCDataModule", context: ts.Context
    ) -> DataLoader:
        """Encrypted predict dataloader.

        Args:
            context (ts.Context): Tenseal encryption context.

        Returns:
            DataLoader: predict dataloader
        """
        enc_csv_predict = EncConvDataset(
            context,
            self.enc_csv_predict,
            self.model._model.conv1.kernel_size,
            self.model._model.conv1.stride[0],
            (32, 32),
        )
        return DataLoader(enc_csv_predict, batch_size=None)

    def teardown(self: "OlindaCDataModule", stage: Optional[str] = None) -> None:
        """Teardown of data module."""
        pass



"""Test encryption route."""

from http import HTTPStatus

from flask import Blueprint, make_response, request, Response
from pydantic import parse_raw_as, ValidationError
import tenseal as ts

from chemxor.schema.test_encryption import TestEncPost, TestEncResponse

test_encryption_bp = Blueprint("test_encryption", __name__, url_prefix="/v1/test")


@test_encryption_bp.post("/")
def test_encryption_post() -> Response:
    """Test encryption.

    Returns:
        Response: Flask response
    """
    # Parse json from requests
    try:
        test_enc = parse_raw_as(TestEncPost, request.json)
    except ValidationError:
        response_body = TestEncResponse(vector="")
        return make_response(response_body.json(), HTTPStatus.BAD_REQUEST)
    except Exception:
        response_body = TestEncResponse(vector="")
        return make_response(response_body.json(), HTTPStatus.INTERNAL_SERVER_ERROR)

    # Perform addition
    context = ts.context_from(bytes.fromhex(test_enc.context))
    new_vector = ts.bfv_vector_from(context, bytes.fromhex(test_enc.vector)) + [
        1,
        1,
        1,
        1,
        1,
    ]
    return make_response(
        TestEncResponse(vector=new_vector.serialize().hex()).json(), HTTPStatus.OK
    )



"""Encrypted Dataset wrapper."""

from typing import Any

import tenseal as ts
from torch.utils.data import Dataset


class EncDataset(Dataset):
    """Encrypted Dataset."""

    def __init__(self: "EncDataset", context: ts.Context, dataset: Dataset) -> None:
        """Encrypted dataset.

        Args:
            context (ts.Context): Tenseal encryption context.
            dataset (Dataset): Initialized dataset class to wrap.

        """
        self.context = context
        self.dataset = dataset

    def __len__(self: "EncDataset") -> int:
        """Length of the dataset.

        Returns:
            int: Length of the dataset.
        """
        return self.dataset.__len__()

    def __getitem__(self: "EncDataset", index: int) -> Any:
        """Get item from the dataset.

        Args:
            index (int): index of the item.

        Returns:
            Any: Item
        """
        items = self.dataset.__getitem__(index)
        # Encrypt items
        enc_item_list = []
        for item in items:
            try:
                _ = self.context.global_scale
                enc_item_list.append(ts.ckks_tensor(self.context, item.reshape(1, -1)))
            except ValueError:
                enc_item_list.append(ts.bfv_tensor(self.context, item.reshape(1, -1)))
        return tuple(enc_item_list)



"""ChemXor."""

from importlib.metadata import PackageNotFoundError, version  # type: ignore
from warnings import filterwarnings

try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"

filterwarnings(action="ignore", category=DeprecationWarning, module="tensorboard")
filterwarnings(action="ignore", category=DeprecationWarning, module="torchvision")



"""Chemxor CLI."""

from pathlib import Path
from typing import Optional

import click

from chemxor.model import (
    FHEOlindaNet,
    FHEOlindaNetOne,
    FHEOlindaNetZero,
    OlindaNet,
    OlindaNetOne,
    OlindaNetZero,
)
from chemxor.service.part_net_client import PartitionNetClient
from chemxor.service.part_net_server import PartitionNetServer
from .. import __version__


@click.command()
@click.argument("model", type=click.Choice(["olinda", "olinda_zero", "olinda_one"]))
@click.option("-cp", "--checkpoint", type=click.Path(exists=True))
@click.option("-o", "--output", default=1)
def serve(model: str, checkpoint: Optional[Path], output: int) -> None:
    """Serve models."""
    if model == "olinda":
        model = OlindaNet(output=output)
        if checkpoint is not None:
            model.load(checkpoint)
        fhe_model = FHEOlindaNet(model)
    elif model == "olinda_zero":
        model = OlindaNetZero(output=output)
        if checkpoint is not None:
            model.load(checkpoint)
        fhe_model = FHEOlindaNetZero(model)
    if model == "olinda_one":
        model = OlindaNetOne(output=output)
        if checkpoint is not None:
            model.load(checkpoint)
        fhe_model = FHEOlindaNetOne(model)
    server = PartitionNetServer(fhe_model)
    server.run()


@click.command()
@click.argument("url")
@click.argument("smile")
def query(url: str, smile: str) -> None:
    """Query models."""
    client = PartitionNetClient(url)
    dec_out = client.query(smile)
    print(f"Model output: {dec_out}")


@click.group()
@click.version_option(version=__version__)
def main() -> None:
    """Chemxor console."""
    pass


main.add_command(serve)
main.add_command(query)



"""MNIST data module."""

from multiprocessing import cpu_count
from pathlib import Path
from typing import Any, Optional, Union

import pytorch_lightning as pl
import tenseal as ts
from torch import nn
from torch.utils.data import DataLoader, random_split
from torchvision import transforms
from torchvision.datasets import MNIST

from chemxor.data_modules.enc_conv_dataset import EncConvDataset
from chemxor.utils import get_project_root_path


# Adapted from https://pytorch-lightning.readthedocs.io/en/stable/extensions/datamodules.html#datamodules
class MNISTDataModule(pl.LightningDataModule):
    """MNIST data module."""

    def __init__(
        self: "MNISTDataModule",
        data_dir: Path = get_project_root_path().joinpath(  # noqa B008
            "data/01_raw/mnist"
        ),
        batch_size: int = 32,
        transform: Optional[Any] = transforms.Compose(  # noqa: B008
            [
                transforms.ToTensor(),  # noqa: B008
                transforms.Normalize((0.1307,), (0.3081,)),  # noqa: B008
            ]
        ),
        target_transform: Optional[Any] = None,
        model: Optional[Union[nn.Module, pl.LightningModule]] = None,
    ) -> None:
        """Initialize.

        Args:
            data_dir (Path): Data directory.
                Defaults to get_project_root_path().joinpath("data/01_raw/mnist").
            batch_size (int): batch size. Defaults to 32.
            transform (Optional[Any]): Tranforms for the inputs.
                Defaults to transforms.Compose([transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))])
            target_transform (Optional[Any]): Transforms for the target. Defaults to None.
            model (Optional[Union[nn.Module, pl.LightningModule]]): Model. Default to None.
        """
        super().__init__()
        self.batch_size = batch_size
        self.data_dir = data_dir
        self.transform = transform
        self.target_transform = target_transform
        self.model = model

    def prepare_data(self: "MNISTDataModule") -> None:
        """Prepare data."""
        # download
        MNIST(self.data_dir, train=True, download=True)
        MNIST(self.data_dir, train=False, download=True)

    def setup(self: "MNISTDataModule", stage: Optional[str] = None) -> None:
        """Setup data module."""
        # Assign train/val datasets for use in dataloaders
        if stage == "fit" or stage is None:
            mnist_full = MNIST(self.data_dir, train=True, transform=self.transform)
            self.mnist_train, self.mnist_val = random_split(mnist_full, [55000, 5000])

        # Assign test dataset for use in dataloader(s)
        if stage == "test" or stage is None:
            self.mnist_test = MNIST(
                self.data_dir, train=False, transform=self.transform
            )

        if stage == "predict" or stage is None:
            self.mnist_predict = MNIST(
                self.data_dir, train=False, transform=self.transform
            )

    def train_dataloader(self: "MNISTDataModule") -> DataLoader:
        """Train dataloader."""
        return DataLoader(
            self.mnist_train, batch_size=self.batch_size, num_workers=cpu_count()
        )

    def enc_train_dataloader(
        self: "MNISTDataModule", context: ts.Context
    ) -> DataLoader:
        """Encrypted train dataloader.

        Args:
            context (ts.Context): Tenseal encryption context.

        Returns:
            DataLoader: train dataloader
        """
        enc_mnist_train = EncConvDataset(
            context,
            self.mnist_train,
            self.model.conv1.kernel_size,
            self.model.conv1.stride[0],
        )
        return DataLoader(enc_mnist_train, batch_size=None)

    def val_dataloader(self: "MNISTDataModule") -> DataLoader:
        """Validation dataloader."""
        return DataLoader(
            self.mnist_val, batch_size=self.batch_size, num_workers=cpu_count()
        )

    def enc_val_dataloader(self: "MNISTDataModule", context: ts.Context) -> DataLoader:
        """Encrypted validation dataloader.

        Args:
            context (ts.Context): Tenseal encryption context.

        Returns:
            DataLoader: validation dataloader
        """
        enc_mnist_val = EncConvDataset(
            context,
            self.mnist_val,
            self.model.conv1.kernel_size,
            self.model.conv1.stride[0],
        )
        return DataLoader(enc_mnist_val, batch_size=None)

    def test_dataloader(self: "MNISTDataModule") -> DataLoader:
        """Test dataloader."""
        return DataLoader(
            self.mnist_test, batch_size=self.batch_size, num_workers=cpu_count()
        )

    def enc_test_dataloader(self: "MNISTDataModule", context: ts.Context) -> DataLoader:
        """Encrypted test dataloader.

        Args:
            context (ts.Context): Tenseal encryption context.

        Returns:
            DataLoader: test dataloader
        """
        enc_mnist_test = EncConvDataset(
            context,
            self.mnist_test,
            self.model.conv1.kernel_size,
            self.model.conv1.stride[0],
        )
        return DataLoader(enc_mnist_test, batch_size=None)

    def predict_dataloader(self: "MNISTDataModule") -> DataLoader:
        """Predict dataloader."""
        return DataLoader(
            self.mnist_predict, batch_size=self.batch_size, num_workers=cpu_count()
        )

    def enc_predict_dataloader(
        self: "MNISTDataModule", context: ts.Context
    ) -> DataLoader:
        """Encrypted predict dataloader.

        Args:
            context (ts.Context): Tenseal encryption context.

        Returns:
            DataLoader: predict dataloader
        """
        enc_mnist_predict = EncConvDataset(
            context,
            self.mnist_predict,
            self.model.conv1.kernel_size,
            self.model.conv1.stride[0],
        )
        return DataLoader(enc_mnist_predict, batch_size=None)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['chemxor',
 'chemxor.cli',
 'chemxor.config',
 'chemxor.data',
 'chemxor.model',
 'chemxor.pipelines',
 'chemxor.routes',
 'chemxor.schema',
 'chemxor.service']

package_data = \
{'': ['*'],
 'chemxor.config': ['base/*', 'base/catalog/*', 'base/parameters/*', 'local/*']}

install_requires = \
['Flask>=2.1.2,<3.0.0',
 'feather-format>=0.4.1,<0.5.0',
 'logzero>=1.7.0,<2.0.0',
 'pandas>=1.4,<2.0',
 'protobuf==3.19.1',
 'pydantic[dotenv]>=1.8.2,<2.0.0',
 'pytorch-lightning>=1.6.1,<2.0.0',
 'rdkit-pypi>=2022.3.1,<2023.0.0',
 'regex==2022.7.9',
 'tenseal>=0.3.11,<0.4.0',
 'torch[cpu]>=1.10.0,<2.0.0']

entry_points = \
{'console_scripts': ['chemxor = chemxor.cli.main:main']}

setup_kwargs = {
    'name': 'chemxor',
    'version': '0.4.0',
    'description': 'Privacy Preserving AI/ML for Drug Discovery',
    'long_description': '# ChemXor\n\nPrivacy Preserving AI/ML for Drug Discovery\n\n---\n\n## Overview\n\nChemXor is an open source library thta provides a set of pre-tuned model architectures for evaluating FHE(Fully homomorphic encryption) encrypted inputs. These models can be trained as normal Pytorch models. It also provides convenient functions to quickly query and host these models as a service with strong privacy guarantees for the end-user. It is built on top of TenSEAL and Pytorch.\n\n### What is Fully Homomorphic Encryption (FHE)?\n\n> A cryptosystem that supports arbitrary computation on ciphertexts is known as fully homomorphic encryption (FHE). Such a scheme enables the construction of programs for any desirable functionality, which can be run on encrypted inputs to produce an encryption of the result. Since such a program need never decrypt its inputs, it can be run by an untrusted party without revealing its inputs and internal state. Fully homomorphic cryptosystems have great practical implications in the outsourcing of private computations. (Wikipedia)\n\n### Why do you need FHE for Machine Learning?\n\nUsing FHE, one can compute on encrypted data, without learning anything about the data. This enables novel privacy preserving interactions between actors in the context of machine learning.\n\n## Tutorials\n\nLook at the notebook directory for tutorials\n\n## Getting Started\n\nChemxor is available on PyPi and can be installed using pip.\n\n```bash\npip install chemxor\n```\n\n### Model selection\n\nAt the moment, one can choose from 3 pre-tuned models.\n\n- OlindaNetZero : Slimmest model with one convolution and 3 linear layers\n- OlindaNet: Model with two convolutions and 4 linear layers\n- OlindaOneNet: Model with four convolutions and 4 linear layers\n\nThese models accept a 32 x 32 input and can be configured to produce a signle or multiple outputs.\n\n```python\nfrom chemxor.models import OlindaNetZero, OlindaNetOne, OlindaNet\n\n# model for regression\nmodel = OlindaNetZero(output = 1)\n```\n\nThe model is a normal Pytorch Lightning module which is compatible with Pytorch NN module.\n\n### Dataset Preparation\n\nChemXor provides two generic Pytorch Lightning Datamodules (Regression, Classification) that can be used to train and evaluate the models. These Datamodules expects raw data as CSV files with two columns (target, SMILES).\n\n```python\nfrom chemxor.data import OlindaCDataModule, OlindaRDataModule\n\ndm_regression = OlindaRDataModule(csv_path="path/to/csv")\n\n# Use the threshold value to automatically create categorical \n# classes from the target column of the CSV\ndm_classification = OlindaCDataModule(csv_path="path/to/csv", threshold=[0.5])\n```\n\nThe DataModules will take care of converting the smiles input to 32 x 32 images.\n\n### Model Training\n\nIt is recommended to use a Pytorch Lightning trainer to train the models. Although a normal Pytorch training loop can also be used.\n\n```python\nimport pytorch_lightning as pl\n\n# Save the best 3 checkpoints based on validation loss\ncheckpoint_callback = pl.callbacks.ModelCheckpoint(\n        dirpath="path/to/save/checkpoints",\n        save_top_k=3,\n        monitor="VAL_Loss",\n    )\ntrainer = pl.Trainer(callbacks=[checkpoint_callback], accelerator="auto")\ntrainer.fit(model=model, datamodule=data_module)\n```\n\n### FHE models\n\nAfter training, the models can be wrapped using their specific FHE wrappers to process FHE inputs. FHE wrappers will take care of Tenseal context parameters and keys management.\n\n```python\nfrom chemxor.models import OlindaNetZero, OlindaNetOne, OlindaNet\nfrom chemxor.models import FHEOlindaNetZero, FHEOlindaNetOne, FHEOlindaNet\n\nmodel = OlindaNetZero(output = 1)\nmodel.load("path/to/checkpoint")\nfhe_model = FHEOlindaNetZero(model=model)\n```\n\n### FHE inputs evaluation\n\nThe Datamodules can generate Pytorch dataloaders that produce encrypted inputs for the model.\n\n```python\nfrom chemxor.data import OlindaCDataModule, OlindaRDataModule\n\ndm_regression = OlindaRDataModule(csv_path="path/to/csv")\ndm_regression.setup("test")\nenc_data_loader = dm_classification.enc_dataloader(context=fhe_model.enc_context)\nenc_sample = next(iter(enc_data_loader))\n```\n\nAlso, the FHE models are partitioned to control multiplicative depth. So, the forward function is modified to accept a step parameter. For testing, The FHE model can be evaluated locally as follows:\n\n```python\nfrom chemxor.utils import prepare_fhe_input\n\noutput = enc_sample\nfor step in range(fhe_model.steps):\n    output = fhe_model(output, step)\n    dec_out = output.decrypt()\n    output = prepare_fhe_input(\n                    dec_out,\n                    fhe_model.pre_process[step],\n                    fhe_model.enc_context\n                )\n\n# final decryted output\ndecrypted_output = output.decrypt()\n```\n\nThis process can automated using a utility function provided by ChemXor\n\n```python\nfrom chemxor.utils import evaluate_fhe_model\n\ndecrypted_output = evaluate_fhe_model(fhe_model, enc_sample, decrypt = True)\n```\n\n### Serve models\n\nFHE Models can be served in the form of a Flask app as follows:\n\n```python\nfrom chemxor.service import PartitionNetServer\n\nfhe_model_server = PartitionNetServer(fhe_model)\n\nif __name__ == "__main__":\n    fhe_model_server.run()\n```\n\nChemXor\'s Pre defined Models can also be served using the CLI\n\n```bash\nchemxor serve olida|olinda_zero|olinda_one \n```\n\n### Query models\n\n```bash\nchemxor query [model url] [molecule SMILES] \n```\n\n## Developing\n\nWe use poetry to manage project dependecies. Use poetry to install project in editable mode.\n\n```bash\npoetry install\n```\n\n## License\n\nThis project is licensed under GNU AFFERO GENERAL PUBLIC LICENSE Version 3.\nThe direct and indirect dependecies of the project are licensed as follows:\n\n| Name                      | Version     | License                                                                                             | Author                                                                                                           |\n|---------------------------|-------------|-----------------------------------------------------------------------------------------------------|------------------------------------------------------------------------------------------------------------------|\n| Babel                     | 2.10.1      | BSD License                                                                                         | Armin Ronacher                                                                                                   |\n| CacheControl              | 0.12.11     | Apache Software License                                                                             | Eric Larson                                                                                                      |\n| Faker                     | 13.12.0     | MIT License                                                                                         | joke2k                                                                                                           |\n| Flask                     | 2.1.2       | BSD License                                                                                         | Armin Ronacher                                                                                                   |\n| GitPython                 | 3.1.27      | BSD License                                                                                         | Sebastian Thiel, Michael Trier                                                                                   |\n| Jinja2                    | 3.1.2       | BSD License                                                                                         | Armin Ronacher                                                                                                   |\n| Markdown                  | 3.3.7       | BSD License                                                                                         | Manfred Stienstra, Yuri takhteyev and Waylan limberg                                                             |\n| MarkupSafe                | 2.1.1       | BSD License                                                                                         | Armin Ronacher                                                                                                   |\n| Pillow                    | 9.1.1       | Historical Permission Notice and Disclaimer (HPND)                                                  | Alex Clark (PIL Fork Author)                                                                                     |\n| PyJWT                     | 2.3.0       | MIT License                                                                                         | Jose Padilla                                                                                                     |\n| PyNaCl                    | 1.4.0       | Apache License 2.0                                                                                  | The PyNaCl developers                                                                                            |\n| PyYAML                    | 5.4.1       | MIT License                                                                                         | Kirill Simonov                                                                                                   |\n| Pygments                  | 2.12.0      | BSD License                                                                                         | Georg Brandl                                                                                                     |\n| SQLAlchemy                | 1.4.36      | MIT License                                                                                         | Mike Bayer                                                                                                       |\n| SecretStorage             | 3.3.2       | BSD License                                                                                         | Dmitry Shachnev                                                                                                  |\n| Send2Trash                | 1.8.0       | BSD License                                                                                         | Andrew Senetar                                                                                                   |\n| Werkzeug                  | 2.1.2       | BSD License                                                                                         | Armin Ronacher                                                                                                   |\n| absl-py                   | 1.0.0       | Apache Software License                                                                             | The Abseil Authors                                                                                               |\n| aiofiles                  | 0.6.0       | Apache Software License                                                                             | Tin Tvrtkovic                                                                                                    |\n| aiohttp                   | 3.8.1       | Apache Software License                                                                             | UNKNOWN                                                                                                          |\n| aiosignal                 | 1.2.0       | Apache Software License                                                                             | Nikolay Kim                                                                                                      |\n| anyconfig                 | 0.10.1      | MIT License                                                                                         | Satoru SATOH                                                                                                     |\n| anyio                     | 3.6.1       | MIT License                                                                                         | Alex Grönholm                                                                                                    |\n| argcomplete               | 1.12.3      | Apache Software License                                                                             | Andrey Kislyuk                                                                                                   |\n| argon2-cffi               | 21.3.0      | MIT License                                                                                         | UNKNOWN                                                                                                          |\n| argon2-cffi-bindings      | 21.2.0      | MIT License                                                                                         | Hynek Schlawack                                                                                                  |\n| arrow                     | 1.2.2       | Apache Software License                                                                             | Chris Smith                                                                                                      |\n| ascii-magic               | 1.6         | MIT License                                                                                         | Leandro Barone                                                                                                   |\n| ase                       | 3.22.1      | GNU Lesser General Public License v2 or later (LGPLv2+)                                             | UNKNOWN                                                                                                          |\n| asgiref                   | 3.5.2       | BSD License                                                                                         | Django Software Foundation                                                                                       |\n| async-timeout             | 4.0.2       | Apache Software License                                                                             | Andrew Svetlov <andrew.svetlov@gmail.com>                                                                        |\n| attrs                     | 21.4.0      | MIT License                                                                                         | Hynek Schlawack                                                                                                  |\n| autodp                    | 0.2         | Apache Software License                                                                             | Yu-Xiang Wang                                                                                                    |\n| backcall                  | 0.2.0       | BSD License                                                                                         | Thomas Kluyver                                                                                                   |\n| backports.cached-property | 1.0.1       | MIT License                                                                                         | Aleksei Stepanov                                                                                                 |\n| bandit                    | 1.7.4       | Apache Software License                                                                             | PyCQA                                                                                                            |\n| bcrypt                    | 3.2.0       | Apache Software License                                                                             | The Python Cryptographic Authority developers                                                                    |\n| beautifulsoup4            | 4.11.1      | MIT License                                                                                         | Leonard Richardson                                                                                               |\n| binaryornot               | 0.4.4       | BSD License                                                                                         | Audrey Roy Greenfeld                                                                                             |\n| black                     | 22.3.0      | MIT License                                                                                         | Łukasz Langa                                                                                                     |\n| bleach                    | 5.0.0       | Apache Software License                                                                             | UNKNOWN                                                                                                          |\n| cachetools                | 4.2.4       | MIT License                                                                                         | Thomas Kemmer                                                                                                    |\n| cachy                     | 0.3.0       | MIT License                                                                                         | Sébastien Eustace                                                                                                |\n| certifi                   | 2022.5.18.1 | Mozilla Public License 2.0 (MPL 2.0)                                                                | Kenneth Reitz                                                                                                    |\n| cffi                      | 1.15.0      | MIT License                                                                                         | Armin Rigo, Maciej Fijalkowski                                                                                   |\n| cfgv                      | 3.3.1       | MIT License                                                                                         | Anthony Sottile                                                                                                  |\n| chardet                   | 4.0.0       | GNU Library or Lesser General Public License (LGPL)                                                 | Mark Pilgrim                                                                                                     |\n| charset-normalizer        | 2.0.12      | MIT License                                                                                         | Ahmed TAHRI @Ousret                                                                                              |\n| chemxor                   | 0.1.0       | GPLv3                                                                     | Ersilia Open Source Initiative                                                                                               |\n| cleo                      | 0.8.1       | MIT License                                                                                         | Sébastien Eustace                                                                                                |\n| click                     | 8.1.3       | BSD License                                                                                         | Armin Ronacher                                                                                                   |\n| clikit                    | 0.6.2       | MIT License                                                                                         | Sébastien Eustace                                                                                                |\n| colorlog                  | 6.6.0       | MIT License                                                                                         | Sam Clements                                                                                                     |\n| cookiecutter              | 1.7.3       | BSD License                                                                                         | Audrey Roy                                                                                                       |\n| coverage                  | 6.4         | Apache Software License                                                                             | Ned Batchelder and 152 others                                                                                    |\n| crashtest                 | 0.3.1       | MIT License                                                                                         | Sébastien Eustace                                                                                                |\n| cryptography              | 37.0.2      | Apache Software License; BSD License                                                                | The Python Cryptographic Authority and individual contributors                                                   |\n| cycler                    | 0.11.0      | BSD License                                                                                         | Thomas A Caswell                                                                                                 |\n| darglint                  | 1.8.1       | MIT License                                                                                         | terrencepreilly                                                                                                  |\n| debugpy                   | 1.6.0       | Eclipse Public License 2.0 (EPL-2.0); MIT License                                                   | Microsoft Corporation                                                                                            |\n| decorator                 | 5.1.1       | BSD License                                                                                         | Michele Simionato                                                                                                |\n| defusedxml                | 0.7.1       | Python Software Foundation License                                                                  | Christian Heimes                                                                                                 |\n| distlib                   | 0.3.4       | Python Software Foundation License                                                                  | Vinay Sajip                                                                                                      |\n| dnspython                 | 2.2.1       | ISC                                                                                                 | Bob Halley                                                                                                       |\n| dparse                    | 0.5.1       | MIT License                                                                                         | Jannis Gebauer                                                                                                   |\n| dynaconf                  | 3.1.8       | MIT License                                                                                         | Bruno Rocha                                                                                                      |\n| email-validator           | 1.2.1       | CC0 1.0 Universal (CC0 1.0) Public Domain Dedication                                                | Joshua Tauberer                                                                                                  |\n| entrypoints               | 0.4         | MIT License                                                                                         | Thomas Kluyver                                                                                                   |\n| eva                       | 1.0.1       | UNKNOWN                                                                                             | Microsoft Research EVA compiler team                                                                             |\n| factory-boy               | 3.2.1       | MIT License                                                                                         | Mark Sandstrom                                                                                                   |\n| fastapi                   | 0.66.1      | MIT License                                                                                         | Sebastián Ramírez                                                                                                |\n| fastjsonschema            | 2.15.3      | BSD License                                                                                         | Michal Horejsek                                                                                                  |\n| filelock                  | 3.7.0       | Public Domain                                                                                       | Benedikt Schmitt                                                                                                 |\n| flake8                    | 4.0.1       | MIT License                                                                                         | Tarek Ziade                                                                                                      |\n| flake8-annotations        | 2.9.0       | MIT License                                                                                         | S Co1                                                                                                            |\n| flake8-bandit             | 3.0.0       | MIT License                                                                                         | Tyler Wince                                                                                                      |\n| flake8-black              | 0.3.3       | MIT License                                                                                         | Peter J. A. Cock                                                                                                 |\n| flake8-bugbear            | 22.4.25     | MIT License                                                                                         | Łukasz Langa                                                                                                     |\n| flake8-docstrings         | 1.6.0       | MIT License                                                                                         | Simon ANDRÉ                                                                                                      |\n| flake8-import-order       | 0.18.1      | GNU Lesser General Public License v3 (LGPLv3); MIT License                                          | Alex Stapleton                                                                                                   |\n| flake8-polyfill           | 1.0.2       | MIT License                                                                                         | Ian Cordasco                                                                                                     |\n| fonttools                 | 4.33.3      | MIT License                                                                                         | Just van Rossum                                                                                                  |\n| forbiddenfruit            | 0.1.4       | GNU General Public License v3 or later (GPLv3+); MIT License                                        | Lincoln de Sousa                                                                                                 |\n| frozenlist                | 1.3.0       | Apache Software License                                                                             | UNKNOWN                                                                                                          |\n| fsspec                    | 2022.1.0    | BSD License                                                                                         | UNKNOWN                                                                                                          |\n| gensim                    | 4.2.0       | LGPL-2.1-only                                                                                       | Radim Rehurek                                                                                                    |\n| gevent                    | 21.8.0      | MIT License                                                                                         | Denis Bilenko                                                                                                    |\n| gitdb                     | 4.0.9       | BSD License                                                                                         | Sebastian Thiel                                                                                                  |\n| google-auth               | 2.6.6       | Apache Software License                                                                             | Google Cloud Platform                                                                                            |\n| google-auth-oauthlib      | 0.4.6       | Apache Software License                                                                             | Google Cloud Platform                                                                                            |\n| graphql-core              | 3.2.1       | MIT License                                                                                         | Christoph Zwerschke                                                                                              |\n| greenlet                  | 1.1.2       | MIT License                                                                                         | Alexey Borzenkov                                                                                                 |\n| griddify                  | 0.0.1       | MIT License                                                                                         | Miquel Duran-Frigola                                                                                             |\n| grpcio                    | 1.46.3      | Apache Software License                                                                             | The gRPC Authors                                                                                                 |\n| h11                       | 0.13.0      | MIT License                                                                                         | Nathaniel J. Smith                                                                                               |\n| html5lib                  | 1.1         | MIT License                                                                                         | UNKNOWN                                                                                                          |\n| httptools                 | 0.4.0       | MIT License                                                                                         | Yury Selivanov                                                                                                   |\n| identify                  | 2.5.1       | MIT License                                                                                         | Chris Kuehl                                                                                                      |\n| idna                      | 3.3         | BSD License                                                                                         | Kim Davies                                                                                                       |\n| importlib-metadata        | 4.11.4      | Apache Software License                                                                             | Jason R. Coombs                                                                                                  |\n| importlib-resources       | 5.7.1       | Apache Software License                                                                             | Barry Warsaw                                                                                                     |\n| iniconfig                 | 1.1.1       | MIT License                                                                                         | Ronny Pfannschmidt, Holger Krekel                                                                                |\n| ipykernel                 | 6.13.0      | BSD License                                                                                         | IPython Development Team                                                                                         |\n| ipython                   | 7.34.0      | BSD License                                                                                         | The IPython Development Team                                                                                     |\n| ipython-genutils          | 0.2.0       | BSD License                                                                                         | IPython Development Team                                                                                         |\n| ipywidgets                | 7.7.0       | BSD License                                                                                         | IPython Development Team                                                                                         |\n| itsdangerous              | 2.1.2       | BSD License                                                                                         | Armin Ronacher                                                                                                   |\n| jedi                      | 0.18.1      | MIT License                                                                                         | David Halter                                                                                                     |\n| jeepney                   | 0.8.0       | MIT License                                                                                         | Thomas Kluyver                                                                                                   |\n| jinja2-time               | 0.2.0       | MIT License                                                                                         | Raphael Pierzina                                                                                                 |\n| jmespath                  | 0.10.0      | MIT License                                                                                         | James Saryerwinnie                                                                                               |\n| joblib                    | 1.1.0       | BSD License                                                                                         | Gael Varoquaux                                                                                                   |\n| json5                     | 0.9.8       | Apache Software License                                                                             | Dirk Pranke                                                                                                      |\n| jsonschema                | 4.5.1       | MIT License                                                                                         | Julian Berman                                                                                                    |\n| jupyter-client            | 7.3.1       | BSD License                                                                                         | Jupyter Development Team                                                                                         |\n| jupyter-core              | 4.10.0      | BSD License                                                                                         | Jupyter Development Team                                                                                         |\n| jupyter-server            | 1.17.0      | BSD License                                                                                         | Jupyter Development Team                                                                                         |\n| jupyterlab                | 3.4.2       | BSD License                                                                                         | Jupyter Development Team                                                                                         |\n| jupyterlab-pygments       | 0.2.2       | BSD                                                                                                 | Jupyter Development Team                                                                                         |\n| jupyterlab-server         | 2.14.0      | BSD License                                                                                         | UNKNOWN                                                                                                          |\n| jupyterlab-widgets        | 1.1.0       | BSD License                                                                                         | Jupyter Development Team                                                                                         |\n| kedro                     | 0.18.1      | Apache Software License (Apache 2.0)                                                                | Kedro                                                                                                            |\n| kedro-viz                 | 4.6.0       | Apache Software License (Apache 2.0)                                                                | Kedro                                                                                                            |\n| keyring                   | 23.5.0      | MIT License; Python Software Foundation License                                                     | Kang Zhang                                                                                                       |\n| kiwisolver                | 1.4.2       | BSD License                                                                                         | UNKNOWN                                                                                                          |\n| lap                       | 0.4.0       | BSD (2-clause)                                                                                      | UNKNOWN                                                                                                          |\n| lark-parser               | 0.12.0      | MIT License                                                                                         | Erez Shinan                                                                                                      |\n| llvmlite                  | 0.38.1      | BSD                                                                                                 | UNKNOWN                                                                                                          |\n| lockfile                  | 0.12.2      | MIT License                                                                                         | OpenStack                                                                                                        |\n| loguru                    | 0.5.3       | MIT License                                                                                         | Delgan                                                                                                           |\n| logzero                   | 1.7.0       | MIT License                                                                                         | Chris Hager                                                                                                      |\n| matplotlib                | 3.5.2       | Python Software Foundation License                                                                  | John D. Hunter, Michael Droettboom                                                                               |\n| matplotlib-inline         | 0.1.3       | BSD 3-Clause                                                                                        | IPython Development Team                                                                                         |\n| mccabe                    | 0.6.1       | MIT License                                                                                         | Ian Cordasco                                                                                                     |\n| mistune                   | 0.8.4       | BSD License                                                                                         | Hsiaoming Yang                                                                                                   |\n| mpmath                    | 1.2.1       | BSD License                                                                                         | Fredrik Johansson                                                                                                |\n| msgpack                   | 1.0.3       | Apache Software License                                                                             | Inada Naoki                                                                                                      |\n| multidict                 | 6.0.2       | Apache Software License                                                                             | Andrew Svetlov                                                                                                   |\n| mypy                      | 0.942       | MIT License                                                                                         | Jukka Lehtosalo                                                                                                  |\n| mypy-extensions           | 0.4.3       | MIT License                                                                                         | The mypy developers                                                                                              |\n| names                     | 0.3.0       | MIT License                                                                                         | Trey Hunner                                                                                                      |\n| nbclassic                 | 0.3.7       | BSD License                                                                                         | Jupyter Development Team                                                                                         |\n| nbclient                  | 0.6.3       | BSD License                                                                                         | Jupyter Development Team                                                                                         |\n| nbconvert                 | 6.5.0       | BSD License                                                                                         | Jupyter Development Team                                                                                         |\n| nbformat                  | 5.4.0       | BSD License                                                                                         | Jupyter Development Team                                                                                         |\n| nest-asyncio              | 1.5.5       | BSD License                                                                                         | Ewald R. de Wit                                                                                                  |\n| networkx                  | 2.8.2       | BSD License                                                                                         | Aric Hagberg                                                                                                     |\n| nglview                   | 3.0.3       | MIT License                                                                                         | Alexander S. Rose, Hai Nguyen                                                                                    |\n| nltk                      | 3.7         | Apache Software License                                                                             | NLTK Team                                                                                                        |\n| nodeenv                   | 1.6.0       | BSD License                                                                                         | Eugene Kalinin                                                                                                   |\n| notebook                  | 6.4.11      | BSD License                                                                                         | Jupyter Development Team                                                                                         |\n| notebook-shim             | 0.1.0       | BSD License                                                                                         | Jupyter Development Team                                                                                         |\n| nox                       | 2022.1.7    | Apache Software License                                                                             | Alethea Katherine Flowers                                                                                        |\n| numba                     | 0.55.2      | BSD License                                                                                         | UNKNOWN                                                                                                          |\n| numpy                     | 1.22.4      | BSD License                                                                                         | Travis E. Oliphant et al.                                                                                        |\n| oauthlib                  | 3.2.0       | BSD License                                                                                         | The OAuthlib Community                                                                                           |\n| onnx                      | 1.11.0      | Apache License v2.0                                                                                 | ONNX                                                                                                             |\n| packaging                 | 21.3        | Apache Software License; BSD License                                                                | Donald Stufft and individual contributors                                                                        |\n| pandas                    | 1.3.4       | BSD License                                                                                         | The Pandas Development Team                                                                                      |\n| pandocfilters             | 1.5.0       | BSD License                                                                                         | John MacFarlane                                                                                                  |\n| parso                     | 0.8.3       | MIT License                                                                                         | David Halter                                                                                                     |\n| pastel                    | 0.2.1       | MIT License                                                                                         | Sébastien Eustace                                                                                                |\n| pathspec                  | 0.9.0       | Mozilla Public License 2.0 (MPL 2.0)                                                                | Caleb P. Burns                                                                                                   |\n| pbr                       | 5.9.0       | Apache Software License                                                                             | OpenStack                                                                                                        |\n| pep517                    | 0.12.0      | MIT License                                                                                         | Thomas Kluyver                                                                                                   |\n| pexpect                   | 4.8.0       | ISC License (ISCL)                                                                                  | Noah Spurrier; Thomas Kluyver; Jeff Quast                                                                        |\n| pickleshare               | 0.7.5       | MIT License                                                                                         | Ville Vainio                                                                                                     |\n| pip-tools                 | 6.6.2       | BSD License                                                                                         | Vincent Driessen                                                                                                 |\n| pkginfo                   | 1.8.2       | MIT License                                                                                         | Tres Seaver, Agendaless Consulting                                                                               |\n| platformdirs              | 2.5.2       | MIT License                                                                                         | UNKNOWN                                                                                                          |\n| plotly                    | 5.8.0       | MIT                                                                                                 | Chris P                                                                                                          |\n| pluggy                    | 1.0.0       | MIT License                                                                                         | Holger Krekel                                                                                                    |\n| poetry                    | 1.1.13      | MIT License                                                                                         | Sébastien Eustace                                                                                                |\n| poetry-core               | 1.0.8       | MIT License                                                                                         | Sébastien Eustace                                                                                                |\n| poyo                      | 0.5.0       | MIT License                                                                                         | Raphael Pierzina                                                                                                 |\n| pre-commit                | 2.19.0      | MIT License                                                                                         | Anthony Sottile                                                                                                  |\n| prometheus-client         | 0.14.1      | Apache Software License                                                                             | Brian Brazil                                                                                                     |\n| prompt-toolkit            | 3.0.29      | BSD License                                                                                         | Jonathan Slenders                                                                                                |\n| protobuf                  | 3.19.1      | 3-Clause BSD License                                                                                | UNKNOWN                                                                                                          |\n| psutil                    | 5.9.1       | BSD License                                                                                         | Giampaolo Rodola                                                                                                 |\n| ptyprocess                | 0.7.0       | ISC License (ISCL)                                                                                  | Thomas Kluyver                                                                                                   |\n| py                        | 1.11.0      | MIT License                                                                                         | holger krekel, Ronny Pfannschmidt, Benjamin Peterson and others                                                  |\n| pyDeprecate               | 0.3.2       | MIT                                                                                                 | Jiri Borovec                                                                                                     |\n| pyarrow                   | 6.0.0       | Apache Software License                                                                             | UNKNOWN                                                                                                          |\n| pyasn1                    | 0.4.8       | BSD License                                                                                         | Ilya Etingof                                                                                                     |\n| pyasn1-modules            | 0.2.8       | BSD License                                                                                         | Ilya Etingof                                                                                                     |\n| pycodestyle               | 2.8.0       | MIT License                                                                                         | Johann C. Rocholl                                                                                                |\n| pycparser                 | 2.21        | BSD License                                                                                         | Eli Bendersky                                                                                                    |\n| pydantic                  | 1.9.1       | MIT License                                                                                         | Samuel Colvin                                                                                                    |\n| pydocstyle                | 6.1.1       | MIT License                                                                                         | Amir Rachum                                                                                                      |\n| pyflakes                  | 2.4.0       | MIT License                                                                                         | A lot of people                                                                                                  |\n| pylev                     | 1.4.0       | BSD License                                                                                         | Daniel Lindsley                                                                                                  |\n| pymbolic                  | 2021.1      | MIT License                                                                                         | Andreas Kloeckner                                                                                                |\n| pynndescent               | 0.5.7       | BSD                                                                                                 | Leland McInnes                                                                                                   |\n| pyparsing                 | 3.0.9       | MIT License                                                                                         | UNKNOWN                                                                                                          |\n| pyrsistent                | 0.18.1      | MIT License                                                                                         | Tobias Gustafsson                                                                                                |\n| pytest                    | 7.1.2       | MIT License                                                                                         | Holger Krekel, Bruno Oliveira, Ronny Pfannschmidt, Floris Bruynooghe, Brianna Laugher, Florian Bruhin and others |\n| pytest-cov                | 3.0.0       | MIT License                                                                                         | Marc Schlaich                                                                                                    |\n| pytest-mock               | 3.7.0       | MIT License                                                                                         | Bruno Oliveira                                                                                                   |\n| python-dateutil           | 2.8.2       | Apache Software License; BSD License                                                                | Gustavo Niemeyer                                                                                                 |\n| python-dotenv             | 0.20.0      | BSD License                                                                                         | Saurabh Kumar                                                                                                    |\n| python-json-logger        | 2.0.2       | BSD License                                                                                         | Zakaria Zajac                                                                                                    |\n| python-multipart          | 0.0.5       | Apache Software License                                                                             | Andrew Dunham                                                                                                    |\n| python-slugify            | 6.1.2       | MIT License                                                                                         | Val Neekman                                                                                                      |\n| pytools                   | 2022.1.7    | MIT License                                                                                         | Andreas Kloeckner                                                                                                |\n| pytorch-lightning         | 1.6.3       | Apache Software License                                                                             | William Falcon et al.                                                                                            |\n| pytz                      | 2022.1      | MIT License                                                                                         | Stuart Bishop                                                                                                    |\n| pyzmq                     | 23.0.0      | BSD License; GNU Library or Lesser General Public License (LGPL)                                    | Brian E. Granger, Min Ragan-Kelley                                                                               |\n| rdkit-pypi                | 2022.3.2.1  | BSD-3-Clause                                                                                        | Christopher Kuenneth                                                                                             |\n| regex                     | 2022.4.24   | Apache Software License                                                                             | Matthew Barnett                                                                                                  |\n| requests                  | 2.27.1      | Apache Software License                                                                             | Kenneth Reitz                                                                                                    |\n| requests-oauthlib         | 1.3.1       | BSD License                                                                                         | Kenneth Reitz                                                                                                    |\n| requests-toolbelt         | 0.9.1       | Apache Software License                                                                             | Ian Cordasco, Cory Benfield                                                                                      |\n| rope                      | 0.21.1      | GNU Lesser General Public License v3 or later (LGPLv3+)                                             | Ali Gholami Rudi                                                                                                 |\n| rsa                       | 4.8         | Apache Software License                                                                             | Sybren A. Stüvel                                                                                                 |\n| safety                    | 1.10.3      | MIT License                                                                                         | pyup.io                                                                                                          |\n| scikit-learn              | 1.1.1       | new BSD                                                                                             | UNKNOWN                                                                                                          |\n| scipy                     | 1.7.3       | BSD License                                                                                         | UNKNOWN                                                                                                          |\n| semver                    | 2.13.0      | BSD License                                                                                         | Kostiantyn Rybnikov                                                                                              |\n| setuptools-scm            | 6.4.2       | MIT License                                                                                         | Ronny Pfannschmidt                                                                                               |\n| shellingham               | 1.4.0       | ISC License (ISCL)                                                                                  | Tzu-ping Chung                                                                                                   |\n| six                       | 1.16.0      | MIT License                                                                                         | Benjamin Peterson                                                                                                |\n| smart-open                | 6.0.0       | MIT License                                                                                         | Radim Rehurek                                                                                                    |\n| smmap                     | 5.0.0       | BSD License                                                                                         | Sebastian Thiel                                                                                                  |\n| sniffio                   | 1.2.0       | Apache Software License; MIT License                                                                | Nathaniel J. Smith                                                                                               |\n| snowballstemmer           | 2.2.0       | BSD License                                                                                         | Snowball Developers                                                                                              |\n| soupsieve                 | 2.3.2.post1 | MIT License                                                                                         | UNKNOWN                                                                                                          |\n| starlette                 | 0.14.2      | BSD License                                                                                         | Tom Christie                                                                                                     |\n| stevedore                 | 3.5.0       | Apache Software License                                                                             | OpenStack                                                                                                        |\n| strawberry-graphql        | 0.114.0     | MIT License                                                                                         | Patrick Arminio                                                                                                  |\n| syft                      | 0.6.0       | Apache-2.0                                                                                          | OpenMined                                                                                                        |\n| sympy                     | 1.9         | BSD License                                                                                         | SymPy development team                                                                                           |\n| tenacity                  | 8.0.1       | Apache Software License                                                                             | Julien Danjou                                                                                                    |\n| tenseal                   | 0.3.11      | Apache-2.0                                                                                          | OpenMined                                                                                                        |\n| tensorboard               | 2.9.0       | Apache Software License                                                                             | Google Inc.                                                                                                      |\n| tensorboard-data-server   | 0.6.1       | Apache Software License                                                                             | Google Inc.                                                                                                      |\n| tensorboard-plugin-wit    | 1.8.1       | Apache 2.0                                                                                          | Google Inc.                                                                                                      |\n| terminado                 | 0.15.0      | BSD License                                                                                         | UNKNOWN                                                                                                          |\n| text-unidecode            | 1.3         | Artistic License; GNU General Public License (GPL); GNU General Public License v2 or later (GPLv2+) | Mikhail Korobov                                                                                                  |\n| threadpoolctl             | 3.1.0       | BSD License                                                                                         | Thomas Moreau                                                                                                    |\n| tinycss2                  | 1.1.1       | BSD License                                                                                         | UNKNOWN                                                                                                          |\n| toml                      | 0.10.2      | MIT License                                                                                         | William Pearson                                                                                                  |\n| tomli                     | 2.0.1       | MIT License                                                                                         | UNKNOWN                                                                                                          |\n| tomlkit                   | 0.10.2      | MIT License                                                                                         | Sébastien Eustace                                                                                                |\n| toposort                  | 1.7         | Apache Software License                                                                             | "Eric V. Smith"                                                                                                  |\n| torch                     | 1.11.0      | BSD License                                                                                         | PyTorch Team                                                                                                     |\n| torch-tb-profiler         | 0.4.0       | BSD License                                                                                         | PyTorch Team                                                                                                     |\n| torchani                  | 2.2         | MIT                                                                                                 | Xiang Gao                                                                                                        |\n| torchmetrics              | 0.8.2       | Apache Software License                                                                             | PyTorchLightning et al.                                                                                          |\n| torchvision               | 0.12.0      | BSD                                                                                                 | PyTorch Core Team                                                                                                |\n| tornado                   | 6.1         | Apache Software License                                                                             | Facebook                                                                                                         |\n| tqdm                      | 4.64.0      | MIT License; Mozilla Public License 2.0 (MPL 2.0)                                                   | UNKNOWN                                                                                                          |\n| traitlets                 | 5.2.1.post0 | BSD License                                                                                         | UNKNOWN                                                                                                          |\n| typeguard                 | 2.13.3      | MIT License                                                                                         | Alex Grönholm                                                                                                    |\n| typing-extensions         | 4.2.0       | Python Software Foundation License                                                                  | UNKNOWN                                                                                                          |\n| umap-learn                | 0.5.3       | BSD                                                                                                 | UNKNOWN                                                                                                          |\n| urllib3                   | 1.26.9      | MIT License                                                                                         | Andrey Petrov                                                                                                    |\n| uvicorn                   | 0.17.6      | BSD License                                                                                         | Tom Christie                                                                                                     |\n| uvloop                    | 0.16.0      | Apache Software License; MIT License                                                                | Yury Selivanov                                                                                                   |\n| virtualenv                | 20.14.1     | MIT License                                                                                         | Bernat Gabor                                                                                                     |\n| watchgod                  | 0.8.2       | MIT License                                                                                         | Samuel Colvin                                                                                                    |\n| wcwidth                   | 0.2.5       | MIT License                                                                                         | Jeff Quast                                                                                                       |\n| webencodings              | 0.5.1       | BSD License                                                                                         | Geoffrey Sneddon                                                                                                 |\n| websocket-client          | 1.3.2       | Apache Software License                                                                             | liris                                                                                                            |\n| websockets                | 10.3        | BSD License                                                                                         | Aymeric Augustin                                                                                                 |\n| widgetsnbextension        | 3.6.0       | BSD License                                                                                         | IPython Development Team                                                                                         |\n| yarl                      | 1.7.2       | Apache Software License                                                                             | Andrew Svetlov                                                                                                   |\n| zipp                      | 3.8.0       | MIT License                                                                                         | Jason R. Coombs                                                                                                  |\n| zope.event                | 4.5.0       | Zope Public License                                                                                 | Zope Foundation and Contributors                                                                                 |\n| zope.interface            | 5.4.0       | Zope Public License                                                                                 | Zope Foundation and Contributors                                                                                 |\n',
    'author': 'Ankur Kumar',
    'author_email': 'ank@leoank.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.9',
}


setup(**setup_kwargs)



"""Partitioned Network Client."""

import json
from typing import Any

from pydantic import parse_obj_as
import requests
import tenseal as ts
from tqdm import tqdm

from chemxor.schema.fhe_model import (
    ModelInfo,
    PartFHEModelQueryGetResponse,
    PartFHEModelQueryPostRequest,
    PartFHEModelQueryPostResponse,
)
from chemxor.utils import prepare_fhe_input, smiles_to_imcol


class PartitionNetClient:
    """Partitioned network client."""

    def __init__(self: "PartitionNetClient", url: str) -> None:
        """Initialize Client.

        Args:
            url (str): URL of the model service
        """
        self.model_url = url
        self.model_info = self.retrieve_model_info(url)
        self.enc_context = self.create_ts_context_from_model(self.model_info)
        self.public_context = self.enc_context.copy()
        self.public_context.make_context_public()

    def create_ts_context_from_model(
        self: "PartitionNetClient", model_info: ModelInfo
    ) -> ts.Context:
        """Create tenseal context from model info."""
        bits_scale = model_info.context_params.bit_scale
        enc_context = ts.context(
            ts.SCHEME_TYPE.CKKS,
            poly_modulus_degree=model_info.context_params.poly_modulus_degree,
            coeff_mod_bit_sizes=model_info.context_params.coeff_mod_bit_sizes,
        )
        enc_context.global_scale = pow(2, bits_scale)
        enc_context.generate_galois_keys()
        return enc_context

    def retrieve_model_info(self: "PartitionNetClient", url: str) -> Any:
        """Retrieve model information to create encryption context and prepare input.

        Args:
            url (str): URL of the model service

        Returns:
            Any: Model Info
        """
        response = requests.get(self.model_url)
        response = parse_obj_as(
            PartFHEModelQueryGetResponse, json.loads(response.content)
        )
        return response.model_info

    def query(self: "PartitionNetClient", x: str) -> Any:
        """Query a FHE model service.

        Args:
            x (str): SMILES string

        Returns:
            Any: Model output
        """
        output = smiles_to_imcol(x, self.enc_context)
        for step in tqdm(range(self.model_info.model_steps + 1)):
            request = PartFHEModelQueryPostRequest(
                ts_context=self.public_context.serialize().hex(),
                model_input=output.serialize().hex(),
                model_step=step,
            )
            response = requests.post(self.model_url, json=request.json())
            if response.status_code == 200:
                response = parse_obj_as(
                    PartFHEModelQueryPostResponse, json.loads(response.content)
                )
                output = ts.ckks_vector_from(
                    self.enc_context, bytes.fromhex(response.model_output)
                )
                dec_out = output.decrypt()
                output = prepare_fhe_input(
                    dec_out, response.preprocess_next_args, self.enc_context
                )
            else:
                raise Exception("Query Failed!")
        return output.decrypt()



"""Olinda regression data module."""

from copy import deepcopy
from pathlib import Path
from typing import Any, Optional, Union

import joblib
import numpy as np
import pandas as pd
import pytorch_lightning as pl
from rdkit import Chem
from rdkit.Chem import AllChem
import tenseal as ts
import torch as t
from torch.utils.data import DataLoader, Dataset, random_split

from chemxor.data.enc_conv_dataset import EncConvDataset
from chemxor.model import OlindaNet, OlindaNetOne, OlindaNetZero
from chemxor.utils import get_project_root_path

project_root_path = get_project_root_path()
default_path = project_root_path.joinpath(
    "data/01_raw/ModelPreds/eos2r5a/ersilia_output.csv"
)


class OlindaRDataset(Dataset):
    """Olinda Regression Dataset."""

    def __init__(
        self: "OlindaRDataset",
        csv_path: Path = default_path,
        transform: Optional[Any] = None,
        target_transform: Optional[Any] = None,
    ) -> None:
        """Olinda Regression Dataset.

        Args:
            csv_path (Path): csv path.
                Defaults to default_path.
            transform (Optional[Any]): Tranforms for the inputs. Defaults to None.
            target_transform (Optional[Any]): Transforms for the target. Defaults to None.

        """
        self.df_full = pd.read_csv(csv_path.absolute())
        self.transform = transform
        self.target_transform = target_transform
        # load the transformer
        self.grid_transformer = joblib.load(
            project_root_path.joinpath("data/06_models/grid_transformer.joblib")
        )

    def __len__(self: "OlindaRDataset") -> int:
        """Length of the dataset.

        Returns:
            int: Length of the dataset.
        """
        return len(self.df_full)

    def __getitem__(self: "OlindaRDataset", index: int) -> Any:
        """Get item from the dataset.

        Args:
            index (int): index of the item.

        Returns:
            Any: Item
        """
        # Extract smile string
        molecule_smile = self.df_full.iloc[index][1]

        # Extract label
        label = t.tensor(self.df_full.iloc[index][0], dtype=t.float).view((1))

        # Create molecule figerprints
        molecule_fp = AllChem.GetMorganFingerprint(
            Chem.MolFromSmiles(molecule_smile),
            radius=3,
            useCounts=True,
            useFeatures=True,
        )

        # Truncate fp to len 1024
        molecule_fp_truncated = np.zeros((1024), np.uint8)
        for idx, v in molecule_fp.GetNonzeroElements().items():
            nidx = idx % 1024
            molecule_fp_truncated[nidx] += int(v)

        # Create input image
        molecule_img = self.grid_transformer.transform([molecule_fp_truncated])[0]
        molecule_img = t.tensor(molecule_img, dtype=t.float).reshape(1, 32, 32)

        # Apply input transforms
        if self.transform:
            molecule_img = self.transform(molecule_img)

        # Apply target transforms
        if self.target_transform:
            label = self.target_transform(label)

        # Return input and target
        return molecule_img, label


class OlindaRDataModule(pl.LightningDataModule):
    """Olinda Regression data module."""

    def __init__(
        self: "OlindaRDataset",
        csv_path: Path = default_path,
        batch_size: int = 32,
        transform: Optional[Any] = None,
        target_transform: Optional[Any] = None,
        model: Optional[Union[OlindaNet, OlindaNetOne, OlindaNetZero]] = None,
    ) -> None:
        """Olinda Regression data module init.

        Args:
            csv_path (Path): csv path.
                Defaults to default_path.
            batch_size (int): batch size. Defaults to 32.
            transform (Optional[Any]): Tranforms for the inputs. Defaults to None.
            target_transform (Optional[Any]): Transforms for the target. Defaults to None.
            model (Optional[Union[OlindaNet, OlindaNetOne, OlindaNetZero]]): Model. Default to None.
        """
        super().__init__()
        self.csv_path = csv_path
        self.batch_size = batch_size
        self.transform = transform
        self.target_transform = target_transform
        self.model = model

    def prepare_data(self: "OlindaRDataModule") -> None:
        """Prepare data."""
        # download, split, etc...
        # only called on 1 GPU/TPU in distributed
        pass

    def setup(self: "OlindaRDataModule", stage: Optional[str]) -> None:
        """Setup dataloaders.

        Args:
            stage (Optional[str]): Optional pipeline state
        """
        self.dataset = OlindaRDataset(
            self.csv_path,
            self.transform,
            self.target_transform,
        )

        # Calculate split
        train_len = int(len(self.dataset) * 0.7)
        validate_len = int(len(self.dataset) * 0.2)
        test_len = int(len(self.dataset) * 0.1)
        others_len = len(self.dataset) - train_len - validate_len - test_len

        self.csv_train, self.csv_val, self.csv_test, _ = random_split(
            self.dataset,
            [train_len, validate_len, test_len, others_len],
            t.Generator().manual_seed(7777),
        )

        # hack for now, please fixme later
        self.enc_csv_train = deepcopy(self.csv_train)
        self.enc_csv_test = deepcopy(self.csv_test)
        self.csv_predict = deepcopy(self.csv_test)
        self.enc_csv_predict = deepcopy(self.csv_test)

    def train_dataloader(self: "OlindaRDataModule") -> DataLoader:
        """Train dataloader.

        Returns:
            DataLoader: train dataloader
        """
        return DataLoader(self.csv_train, batch_size=self.batch_size)

    def enc_train_dataloader(
        self: "OlindaRDataModule", context: ts.Context
    ) -> DataLoader:
        """Encrypted train dataloader.

        Args:
            context (ts.Context): Tenseal encryption context.

        Returns:
            DataLoader: train dataloader
        """
        enc_csv_train = EncConvDataset(
            context,
            self.enc_csv_train,
            self.model._model.conv1.kernel_size,
            self.model._model.conv1.stride[0],
            (32, 32),
        )
        return DataLoader(enc_csv_train, batch_size=None)

    def val_dataloader(self: "OlindaRDataModule") -> DataLoader:
        """Val dataloader.

        Returns:
            DataLoader: val dataloader
        """
        return DataLoader(self.csv_val, batch_size=self.batch_size)

    def enc_val_dataloader(
        self: "OlindaRDataModule", context: ts.Context
    ) -> DataLoader:
        """Encrypted val dataloader.

        Args:
            context (ts.Context): Tenseal encryption context.

        Returns:
            DataLoader: val dataloader
        """
        enc_csv_val = EncConvDataset(
            context,
            self.enc_csv_val,
            self.model._model.conv1.kernel_size,
            self.model._model.conv1.stride[0],
            (32, 32),
        )
        return DataLoader(enc_csv_val, batch_size=None)

    def test_dataloader(self: "OlindaRDataModule") -> DataLoader:
        """Test data loader.

        Returns:
            DataLoader : test dataloader
        """
        return DataLoader(self.csv_test, batch_size=self.batch_size)

    def enc_test_dataloader(
        self: "OlindaRDataModule", context: ts.Context
    ) -> DataLoader:
        """Encrypted test dataloader.

        Args:
            context (ts.Context): Tenseal encryption context.

        Returns:
            DataLoader: test dataloader
        """
        enc_csv_test = EncConvDataset(
            context,
            self.enc_csv_test,
            self.model._model.conv1.kernel_size,
            self.model._model.conv1.stride[0],
            (32, 32),
        )
        return DataLoader(enc_csv_test, batch_size=None)

    def predict_dataloader(self: "OlindaRDataModule") -> DataLoader:
        """Predict data loader.

        Returns:
            DataLoader : predict dataloader
        """
        return DataLoader(self.csv_predict, batch_size=self.batch_size)

    def enc_predict_dataloader(
        self: "OlindaRDataModule", context: ts.Context
    ) -> DataLoader:
        """Encrypted predict dataloader.

        Args:
            context (ts.Context): Tenseal encryption context.

        Returns:
            DataLoader: predict dataloader
        """
        enc_csv_predict = EncConvDataset(
            context,
            self.enc_csv_predict,
            self.model._model.conv1.kernel_size,
            self.model._model.conv1.stride[0],
            (32, 32),
        )
        return DataLoader(enc_csv_predict, batch_size=None)

    def teardown(self: "OlindaRDataModule", stage: Optional[str] = None) -> None:
        """Teardown of data module."""
        pass
